import torch
from .Criterion import Criterion
from .utils import clear


class L1Cost(Criterion):

    def __init__(self):
        super(L1Cost, self).__init__()
        self.output_tensor = torch.Tensor(1)

    def updateOutput(self, input, target=None):
        assert target is None
        if self.output_tensor is None:
            self.output_tensor = input.new(1)
        self._backend.L1Cost_updateOutput(
            self._backend.library_state,
            input,
            self.output_tensor
        )
        self.output = self.output_tensor[0]
        return self.output

    def updateGradInput(self, input, target=None):
        assert target is None
        self._backend.L1Cost_updateGradInput(
            self._backend.library_state,
            input,
            None,
            self.gradInput
        )
        return self.gradInput

    def clearState(self):
        clear(self, 'output_tensor')
        return super(L1Cost, self).clearState()



from numbers import Integral
import torch
from torch.autograd.function import Function
from torch._thnn import type2backend

from . import _all_functions
from ...modules.utils import _single, _pair, _triple

import warnings


def _check_size_scale_factor(size, scale_factor):
    if size is None and scale_factor is None:
        raise ValueError('either size or scale_factor should be defined')
    if scale_factor is not None and not isinstance(scale_factor, (Integral, tuple)):
        raise ValueError('scale_factor must be of integer type or a tuple of integer types')


def _check_linear_scale_factor(scale_factor, dim=2):
    if dim == 1:
        scale_factor = _single(scale_factor)
    elif dim == 2:
        scale_factor = _pair(scale_factor)
    elif dim == 3:
        scale_factor = _triple(scale_factor)
    else:
        raise ValueError("dim has to be 1, 2 or 3")

    try:
        assert len(scale_factor) == 1 or len(scale_factor) == 2 or len(scale_factor) == 3
        assert all(isinstance(s, Integral) and s >= 1 for s in scale_factor)
    except AssertionError as e:
        raise ValueError('scale_factor must be a non-negative integer, '
                         'or a tuple of non-negative integers for linear, bilinear and trilinear upsampling, but got: '
                         '{}'.format(scale_factor))
    return scale_factor


class UpsamplingNearest1d(Function):

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 3

        _check_size_scale_factor(size, scale_factor)

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None and not isinstance(ctx.scale_factor, Integral):
            raise ValueError('scale_factor must be a single Integer value for nearest neighbor sampling')

        if ctx.scale_factor is None:
            if (ctx.size[0] % input.size(2) != 0):
                raise RuntimeError("output size specified in UpsamplingNearest "
                                   "({}) has to be divisible by the input size, but got: "
                                   "{}".format('x'.join(map(str, ctx.size)),
                                               'x'.join(map(str, input.size()))))
            ctx.scale_factor = ctx.size[0] // input.size(2)

        output = input.new()
        backend = type2backend[type(input)]
        ctx.save_for_backward(input)
        backend.TemporalUpSamplingNearest_updateOutput(
            backend.library_state,
            input,
            output,
            ctx.scale_factor
        )
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_variables
        grad_input = UpsamplingNearest1dBackward.apply(input, grad_output, ctx.scale_factor)
        return grad_input, None, None


class UpsamplingNearest1dBackward(Function):

    @staticmethod
    def forward(ctx, input, grad_output, scale_factor):
        assert grad_output.dim() == 3
        ctx.scale_factor = scale_factor

        grad_input = grad_output.new()
        backend = type2backend[type(input)]
        backend.TemporalUpSamplingNearest_updateGradInput(
            backend.library_state,
            input,
            grad_output,
            grad_input,
            ctx.scale_factor
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        gI = None
        ggO = UpsamplingNearest1d.apply(ggI, None, ctx.scale_factor)

        return gI, ggO, None


class UpsamplingLinear1d(Function):

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 3

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None:
            ctx.scale_factor = _check_linear_scale_factor(ctx.scale_factor, dim=1)

        if ctx.scale_factor is not None:
            ctx.output_size = (
                input.size(2) * ctx.scale_factor[0],
            )
        else:
            ctx.output_size = ctx.size

        ctx.input_size = input.size()
        output = input.new()
        backend = type2backend[type(input)]
        backend.TemporalUpSamplingLinear_updateOutput(
            backend.library_state,
            input,
            output,
            ctx.output_size[0]
        )
        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = UpsamplingLinear1dBackward.apply(grad_output, ctx.input_size, ctx.output_size)
        return grad_input, None, None


class UpsamplingLinear1dBackward(Function):

    @staticmethod
    def forward(ctx, grad_output, input_size, output_size):
        assert grad_output.dim() == 3

        ctx.input_size = input_size
        ctx.output_size = output_size

        grad_output = grad_output.contiguous()
        grad_input = grad_output.new()
        backend = type2backend[type(grad_output)]
        backend.TemporalUpSamplingLinear_updateGradInput(
            backend.library_state,
            grad_output,
            grad_input,
            ctx.input_size[0],
            ctx.input_size[1],
            ctx.input_size[2],
            ctx.output_size[0],
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        ggO = UpsamplingLinear1d.apply(ggI, ctx.output_size, None)

        return ggO, None, None


class UpsamplingNearest2d(Function):

    @staticmethod
    def symbolic(g, input, size=None, scale_factor=None):
        if scale_factor is None:
            scale_factor = 1.0
        if size is not None and set(size) != set([None]):
            warnings.warn("ONNX export failed on UpsamplingNearest2d because size is not supported")
        return g.op("ResizeNearest", input, width_scale_f=scale_factor, height_scale_f=scale_factor)

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 4

        _check_size_scale_factor(size, scale_factor)

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None and not isinstance(ctx.scale_factor, Integral):
            raise ValueError('scale_factor must be a single Integer value for nearest neighbor sampling')

        if ctx.scale_factor is None:
            if (ctx.size[0] % input.size(2) != 0 or
                    ctx.size[1] % input.size(3) != 0):
                raise RuntimeError("output size specified in UpsamplingNearest "
                                   "({}) has to be divisible by the input size, but got: "
                                   "{}".format('x'.join(map(str, ctx.size)),
                                               'x'.join(map(str, input.size()))))
            ctx.scale_factor = ctx.size[0] // input.size(2)
            if ctx.scale_factor != ctx.size[1] // input.size(3):
                raise RuntimeError("input aspect ratio doesn't match the "
                                   "output ratio")

        output = input.new()
        backend = type2backend[type(input)]
        ctx.save_for_backward(input)
        backend.SpatialUpSamplingNearest_updateOutput(
            backend.library_state,
            input,
            output,
            ctx.scale_factor
        )
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_variables
        grad_input = UpsamplingNearest2dBackward.apply(input, grad_output, ctx.scale_factor)
        return grad_input, None, None


class UpsamplingNearest2dBackward(Function):

    @staticmethod
    def forward(ctx, input, grad_output, scale_factor):
        assert grad_output.dim() == 4
        ctx.scale_factor = scale_factor

        grad_input = grad_output.new()
        backend = type2backend[type(input)]
        backend.SpatialUpSamplingNearest_updateGradInput(
            backend.library_state,
            input,
            grad_output,
            grad_input,
            ctx.scale_factor
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        gI = None
        ggO = UpsamplingNearest2d.apply(ggI, None, ctx.scale_factor)

        return gI, ggO, None


class UpsamplingBilinear2d(Function):

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 4

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None:
            ctx.scale_factor = _check_linear_scale_factor(ctx.scale_factor, dim=2)

        if ctx.scale_factor is not None:
            ctx.output_size = (
                input.size(2) * ctx.scale_factor[0],
                input.size(3) * ctx.scale_factor[1],
            )
        else:
            ctx.output_size = ctx.size

        ctx.input_size = input.size()
        output = input.new()
        backend = type2backend[type(input)]
        backend.SpatialUpSamplingBilinear_updateOutput(
            backend.library_state,
            input,
            output,
            ctx.output_size[0],
            ctx.output_size[1],
        )
        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = UpsamplingBilinear2dBackward.apply(grad_output, ctx.input_size, ctx.output_size)
        return grad_input, None, None


class UpsamplingBilinear2dBackward(Function):

    @staticmethod
    def forward(ctx, grad_output, input_size, output_size):
        assert grad_output.dim() == 4

        ctx.input_size = input_size
        ctx.output_size = output_size

        grad_output = grad_output.contiguous()
        grad_input = grad_output.new()
        backend = type2backend[type(grad_output)]
        backend.SpatialUpSamplingBilinear_updateGradInput(
            backend.library_state,
            grad_output,
            grad_input,
            ctx.input_size[0],
            ctx.input_size[1],
            ctx.input_size[2],
            ctx.input_size[3],
            ctx.output_size[0],
            ctx.output_size[1],
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        ggO = UpsamplingBilinear2d.apply(ggI, ctx.output_size, None)

        return ggO, None, None


class UpsamplingNearest3d(Function):

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 5

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None and not isinstance(ctx.scale_factor, Integral):
            raise ValueError('scale_factor must be a single Integer value for nearest neighbor sampling')

        if ctx.scale_factor is None:
            if (ctx.size[0] % input.size(2) != 0 or ctx.size[1] % input.size(3) != 0 or
               ctx.size[2] % input.size(4) != 0):
                raise RuntimeError("output size specified in UpSamplingNearest "
                                   "({}) has to be divisible by the input size, but got: "
                                   "{}".format('x'.join(map(str, ctx.size)),
                                               'x'.join(map(str, input.size()))))
            ctx.scale_factor = ctx.size[0] // input.size(2)
            if (ctx.scale_factor != ctx.size[1] // input.size(3) or
               ctx.scale_factor != ctx.size[2] // input.size(4)):
                raise RuntimeError("input aspect ratio doesn't match the "
                                   "output ratio")

        output = input.new()
        backend = type2backend[type(input)]
        ctx.save_for_backward(input)
        backend.VolumetricUpSamplingNearest_updateOutput(backend.library_state,
                                                         input,
                                                         output,
                                                         ctx.scale_factor)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_variables

        grad_input = UpsamplingNearest3dBackward.apply(input, grad_output, ctx.scale_factor)
        return grad_input, None, None


class UpsamplingNearest3dBackward(Function):

    @staticmethod
    def forward(ctx, input, grad_output, scale_factor):
        assert grad_output.dim() == 5

        ctx.scale_factor = scale_factor
        grad_input = grad_output.new()
        backend = type2backend[type(input)]
        backend.VolumetricUpSamplingNearest_updateGradInput(backend.library_state,
                                                            input,
                                                            grad_output,
                                                            grad_input,
                                                            ctx.scale_factor)
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        gI = None
        ggO = UpsamplingNearest3d.apply(ggI, None, ctx.scale_factor)

        return gI, ggO, None


class UpsamplingTrilinear3d(Function):

    @staticmethod
    def forward(ctx, input, size=None, scale_factor=None):
        assert input.dim() == 5

        ctx.size = size
        ctx.scale_factor = scale_factor

        if ctx.scale_factor is not None:
            ctx.scale_factor = _check_linear_scale_factor(ctx.scale_factor, dim=3)

        if ctx.scale_factor is not None:
            ctx.output_size = (
                input.size(2) * ctx.scale_factor[0],
                input.size(3) * ctx.scale_factor[1],
                input.size(4) * ctx.scale_factor[2],
            )
        else:
            ctx.output_size = ctx.size

        ctx.input_size = input.size()
        output = input.new()
        backend = type2backend[type(input)]
        backend.VolumetricUpSamplingTrilinear_updateOutput(
            backend.library_state,
            input,
            output,
            ctx.output_size[0],
            ctx.output_size[1],
            ctx.output_size[2]
        )
        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = UpsamplingTrilinear3dBackward.apply(grad_output, ctx.input_size, ctx.output_size)
        return grad_input, None, None


class UpsamplingTrilinear3dBackward(Function):

    @staticmethod
    def forward(ctx, grad_output, input_size, output_size):
        assert grad_output.dim() == 5

        ctx.input_size = input_size
        ctx.output_size = output_size
        grad_output = grad_output.contiguous()
        grad_input = grad_output.new()
        backend = type2backend[type(grad_output)]
        backend.VolumetricUpSamplingTrilinear_updateGradInput(
            backend.library_state,
            grad_output,
            grad_input,
            ctx.input_size[0],
            ctx.input_size[1],
            ctx.input_size[2],
            ctx.input_size[3],
            ctx.input_size[4],
            ctx.output_size[0],
            ctx.output_size[1],
            ctx.output_size[2]
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        ggO = UpsamplingTrilinear3d.apply(ggI, ctx.output_size, None)

        return ggO, None, None


_all_functions.append(UpsamplingNearest1d)
_all_functions.append(UpsamplingNearest1dBackward)
_all_functions.append(UpsamplingLinear1d)
_all_functions.append(UpsamplingLinear1dBackward)
_all_functions.append(UpsamplingNearest2d)
_all_functions.append(UpsamplingNearest2dBackward)
_all_functions.append(UpsamplingBilinear2d)
_all_functions.append(UpsamplingBilinear2dBackward)
_all_functions.append(UpsamplingNearest3d)
_all_functions.append(UpsamplingNearest3dBackward)
_all_functions.append(UpsamplingTrilinear3d)
_all_functions.append(UpsamplingTrilinear3dBackward)



import torch


def lbfgs(opfunc, x, config, state=None):
    """
    An implementation of L-BFGS, heavily inspired by minFunc (Mark Schmidt)
    This implementation of L-BFGS relies on a user-provided line
    search function (state.lineSearch). If this function is not
    provided, then a simple learningRate is used to produce fixed
    size steps. Fixed size steps are much less costly than line
    searches, and can be useful for stochastic problems.
    The learning rate is used even when a line search is provided.
    This is also useful for large-scale stochastic problems, where
    opfunc is a noisy approximation of f(x). In that case, the learning
    rate allows a reduction of confidence in the step size.

    Args:
    - `opfunc` : a function that takes a single input (X), the point of
             evaluation, and returns f(X) and df/dX
    - `x` : the initial point
    - `state` : a table describing the state of the optimizer; after each
             call the state is modified
    - `state.maxIter` : Maximum number of iterations allowed
    - `state.maxEval` : Maximum number of function evaluations
    - `state.tolFun` : Termination tolerance on the first-order optimality
    - `state.tolX` : Termination tol on progress in terms of func/param changes
    - `state.lineSearch` : A line search function
    - `state.learningRate` : If no line search provided, then a fixed step size is used

    Returns:
    - `x*` : the new `x` vector, at the optimal point
    - `f`  : a table of all function values:
         `f[1]` is the value of the function before any optimization and
         `f[#f]` is the final fully optimized value, at `x*`

    (Clement Farabet, 2012)
    """

    # (0) get/update state
    if config is None and state is None:
        raise ValueError("lbfgs requires a dictionary to retain state between iterations")
    state = state if state is not None else config
    maxIter = config.get('maxIter', 20)
    maxEval = config.get('maxEval', maxIter * 1.25)
    tolFun = config.get('tolFun', 1e-5)
    tolX = config.get('tolX', 1e-9)
    nCorrection = config.get('nCorrection', 100)
    lineSearch = config.get('lineSearch')
    lineSearchOptions = config.get('lineSearchOptions')
    learningRate = config.get('learningRate', 1)
    isverbose = config.get('verbose', False)

    state.setdefault('funcEval', 0)
    state.setdefault('nIter', 0)

    # verbose function
    if isverbose:
        def verbose(*args):
            args = ('<optim.lbfgs>',) + args
            print(args)
    else:
        def verbose(*args):
            pass

    # evaluate initial f(x) and df/dx
    f, g = opfunc(x)
    f_hist = [f]
    currentFuncEval = 1
    state['funcEval'] += 1
    p = g.size(0)

    # check optimality of initial point
    if 'tmp1' not in state:
        state['tmp1'] = g.new(g.size()).zero_()
    tmp1 = state['tmp1']
    tmp1.copy_(g).abs_()
    if tmp1.sum() <= tolFun:
        verbose('optimality condition below tolFun')
        return x, f_hist

    if 'dir_bufs' not in state:
        # reusable buffers for y's and s's, and their histories
        verbose('creating recyclable direction/step/history buffers')
        state['dir_bufs'] = list(g.new(nCorrection + 1, p).split(1))
        state['stp_bufs'] = list(g.new(nCorrection + 1, p).split(1))
        for i in range(len(state['dir_bufs'])):
            state['dir_bufs'][i] = state['dir_bufs'][i].squeeze(0)
            state['stp_bufs'][i] = state['stp_bufs'][i].squeeze(0)

    # variables cached in state (for tracing)
    d = state.get('d')
    t = state.get('t')
    old_dirs = state.get('old_dirs')
    old_stps = state.get('old_stps')
    Hdiag = state.get('Hdiag')
    g_old = state.get('g_old')
    f_old = state.get('f_old')

    # optimize for a max of maxIter iterations
    nIter = 0
    while nIter < maxIter:
        # keep track of nb of iterations
        nIter += 1
        state['nIter'] += 1

        ############################################################
        # compute gradient descent direction
        ############################################################
        if state['nIter'] == 1:
            d = g.neg()
            old_dirs = []
            old_stps = []
            Hdiag = 1
        else:
            # do lbfgs update (update memory)
            y = state['dir_bufs'].pop()
            s = state['stp_bufs'].pop()
            torch.add(g, -1, g_old, out=y)
            torch.mul(d, t, out=s)
            ys = y.dot(s)  # y*s
            if ys > 1e-10:
                # updating memory
                if len(old_dirs) == nCorrection:
                    # shift history by one (limited-memory)
                    state['dir_bufs'].append(old_dirs.pop(0))
                    state['stp_bufs'].append(old_stps.pop(0))

                # store new direction/step
                old_dirs.append(s)
                old_stps.append(y)

                # update scale of initial Hessian approximation
                Hdiag = ys / y.dot(y)  # (y*y)
            else:
                # put y and s back into the buffer pool
                state['dir_bufs'].append(y)
                state['stp_bufs'].append(s)

            # compute the approximate (L-BFGS) inverse Hessian
            # multiplied by the gradient
            k = len(old_dirs)

            # need to be accessed element-by-element, so don't re-type tensor:
            if 'ro' not in state:
                state['ro'] = torch.Tensor(nCorrection)
            ro = state['ro']

            for i in range(k):
                ro[i] = 1 / old_stps[i].dot(old_dirs[i])

            # iteration in L-BFGS loop collapsed to use just one buffer
            q = tmp1  # reuse tmp1 for the q buffer
            # need to be accessed element-by-element, so don't re-type tensor:
            if 'al' not in state:
                state['al'] = torch.zeros(nCorrection)
            al = state['al']

            torch.mul(g, -1, out=q)
            for i in range(k - 1, -1, -1):
                al[i] = old_dirs[i].dot(q) * ro[i]
                q.add_(-al[i], old_stps[i])

            # multiply by initial Hessian
            r = d  # share the same buffer, since we don't need the old d
            torch.mul(q, Hdiag, out=r)
            for i in range(k):
                be_i = old_stps[i].dot(r) * ro[i]
                r.add_(al[i] - be_i, old_dirs[i])
            # final direction is in r/d (same object)
        if g_old is None:
            g_old = g.clone()
        else:
            g_old.copy_(g)
        f_old = f

        ############################################################
        # compute step length
        ############################################################
        # directional derivative
        gtd = g.dot(d)  # g * d

        # reset initial guess for step size
        if state['nIter'] == 1:
            tmp1.copy_(g).abs_()
            t = min(1, 1 / tmp1.sum()) * learningRate
        else:
            t = learningRate

        # optional line search: user function
        lsFuncEval = 0
        if lineSearch is not None:
            # perform line search, using user function
            f, g, x, t, lsFuncEval = lineSearch(opfunc, x, t, d, f, g, gtd, lineSearchOpts)
            f_hist.append(f)
        else:
            # no line search, simply move with fixed-step
            x.add_(t, d)
            if nIter != maxIter:
                # re-evaluate function only if not in last iteration
                # the reason we do this: in a stochastic setting,
                # no use to re-evaluate that function here
                f, g = opfunc(x)
                lsFuncEval = 1
                f_hist.append(f)

        # update func eval
        currentFuncEval += lsFuncEval
        state['funcEval'] += lsFuncEval

        ############################################################
        # check conditions
        ############################################################
        if nIter == maxIter:
            # no use to run tests
            verbose('reached max number of iterations')
            break

        if currentFuncEval >= maxEval:
            # max nb of function evals
            verbose('max nb of function evals')
            break

        tmp1.copy_(g).abs_()
        if tmp1.sum() <= tolFun:
            # check optimality
            verbose('optimality condition below tolFun')
            break

        # check that progress can be made along that direction
        if gtd > -tolX:
            break

        tmp1.copy_(d).mul_(t).abs_()
        if tmp1.sum() <= tolX:
            # step size below tolX
            verbose('step size below tolX')
            break

        if abs(f - f_old) < tolX:
            # function value changing less than tolX
            verbose('function value changing less than tolX')
            break

    # save state
    state['old_dirs'] = old_dirs
    state['old_stps'] = old_stps
    state['Hdiag'] = Hdiag
    state['g_old'] = g_old
    state['f_old'] = f_old
    state['t'] = t
    state['d'] = d

    # return optimal x, and history of f(x)
    return x, f_hist, currentFuncEval



import torch
import numpy as np
from test_autograd import _make_cov
from torch.autograd import Variable
from common import TestCase, run_tests, skipIfNoLapack

from torch.autograd._functions.linalg import Potrf


class TestPotrf(TestCase):

    def _calc_deriv_numeric(self, A, L, upper):
        # numerical forward derivative
        dA = Variable(_make_cov(5))
        eps = 1e-6
        outb = Potrf.apply(A + (eps / 2) * dA, upper)
        outa = Potrf.apply(A - (eps / 2) * dA, upper)
        dL = (outb - outa) / eps

        return dA, dL

    def _calc_deriv_sym(self, A, L, upper):
        # reverse mode
        Lbar = Variable(torch.rand(5, 5).tril())
        if upper:
            Lbar = Lbar.t()
        L.backward(Lbar)
        Abar = A.grad

        return Abar, Lbar

    def _check_total_variation(self, A, L, upper):
        dA, dL = self._calc_deriv_numeric(A, L, upper)
        Abar, Lbar = self._calc_deriv_sym(A, L, upper)

        # compare df = Tr(dA^T Abar) = Tr(dL^T Lbar)
        df1 = (dL * Lbar).sum()
        df2 = (dA * Abar).sum()

        atol = 1e-5
        rtol = 1e-3
        assert (df1 - df2).abs().data[0] <= atol + rtol * df1.abs().data[0]

    @skipIfNoLapack
    def test_potrf(self):
        for upper in [True, False]:
            A = Variable(_make_cov(5), requires_grad=True)
            L = Potrf.apply(A, upper)
            self._check_total_variation(A, L, upper)

if __name__ == '__main__':
    run_tests()



import torch
from .Container import Container


class Concat(Container):

    def __init__(self, dimension):
        super(Concat, self).__init__()
        self.outputSize = torch.Size()
        self.dimension = dimension

    def updateOutput(self, input):
        outs = []
        for i in range(len(self.modules)):
            currentOutput = self.modules[i].updateOutput(input)
            outs.append(currentOutput)
            if i == 0:
                size = list(currentOutput.size())
            else:
                size[self.dimension] += currentOutput.size(self.dimension)
        self.outputSize = torch.Size(size)
        self.output.resize_(self.outputSize)

        offset = 0
        for i, module in enumerate(self.modules):
            currentOutput = outs[i]
            self.output.narrow(self.dimension, offset, currentOutput.size(self.dimension)).copy_(currentOutput)
            offset = offset + currentOutput.size(self.dimension)

        return self.output

    def updateGradInput(self, input, gradOutput):
        self.gradInput.resize_as_(input)

        offset = 0
        for i, module in enumerate(self.modules):
            currentOutput = module.output
            currentGradInput = module.updateGradInput(input, gradOutput.narrow(
                self.dimension, offset, currentOutput.size(self.dimension)))

            # if the module does not produce a gradInput (for example first layer),: ignore it and move on.
            if currentGradInput:
                if i == 0:
                    self.gradInput.copy_(currentGradInput)
                else:
                    self.gradInput.add_(currentGradInput)

            offset = offset + currentOutput.size(self.dimension)

        return self.gradInput

    def accGradParameters(self, input, gradOutput, scale=1):
        offset = 0
        for i, module in enumerate(self.modules):
            currentOutput = module.output
            module.accGradParameters(
                input,
                gradOutput.narrow(self.dimension, offset, currentOutput.size(self.dimension)),
                scale)
            offset = offset + currentOutput.size(self.dimension)

    def backward(self, input, gradOutput, scale=1):
        self.gradInput.resize_as_(input)
        offset = 0
        for i, module in enumerate(self.modules):
            currentOutput = module.output
            currentGradInput = module.backward(input, gradOutput.narrow(
                self.dimension, offset, currentOutput.size(self.dimension)), scale)
            # if the module.es not produce a gradInput (for example first layer),: ignore it and move on.
            if currentGradInput is not None:
                if i == 0:
                    self.gradInput.copy_(currentGradInput)
                else:
                    self.gradInput.add_(currentGradInput)
            offset = offset + currentOutput.size(self.dimension)

        return self.gradInput

    def accUpdateGradParameters(self, input, gradOutput, lr):
        offset = 0
        for i, module in enumerate(self.modules):
            currentOutput = module.output
            module.accUpdateGradParameters(
                input,
                gradOutput.narrow(self.dimension, offset, currentOutput.size(self.dimension)),
                lr)
            offset = offset + currentOutput.size(self.dimension)

    def __tostring__(self):
        tab = '  '
        line = '\n'
        next = '  |`-> '
        ext = '  |    '
        extlast = '       '
        last = '   +. -> '
        res = torch.type(self)
        res += ' {' + line + tab + 'input'
        for i in range(len(self.modules)):
            if i == len(self.modules) - 1:
                res += line + tab + next + '(' + i + '): ' + str(self.modules[i]).replace(line, line + tab + extlast)
            else:
                res += line + tab + next + '(' + i + '): ' + str(self.modules[i]).replace(line, line + tab + ext)

        res += line + tab + last + 'output'
        res += line + '}'
        return res



import heapq
from torch.autograd import Variable


class Trainer(object):

    def __init__(self, model=None, criterion=None, optimizer=None, dataset=None):
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.dataset = dataset
        self.iterations = 0
        self.stats = {}
        self.plugin_queues = {
            'iteration': [],
            'epoch': [],
            'batch': [],
            'update': [],
        }

    def register_plugin(self, plugin):
        plugin.register(self)

        intervals = plugin.trigger_interval
        if not isinstance(intervals, list):
            intervals = [intervals]
        for duration, unit in intervals:
            queue = self.plugin_queues[unit]
            queue.append((duration, len(queue), plugin))

    def call_plugins(self, queue_name, time, *args):
        args = (time,) + args
        queue = self.plugin_queues[queue_name]
        if len(queue) == 0:
            return
        while queue[0][0] <= time:
            plugin = queue[0][2]
            getattr(plugin, queue_name)(*args)
            for trigger in plugin.trigger_interval:
                if trigger[1] == queue_name:
                    interval = trigger[0]
            new_item = (time + interval, queue[0][1], plugin)
            heapq.heappushpop(queue, new_item)

    def run(self, epochs=1):
        for q in self.plugin_queues.values():
            heapq.heapify(q)

        for i in range(1, epochs + 1):
            self.train()
            self.call_plugins('epoch', i)

    def train(self):
        for i, data in enumerate(self.dataset, self.iterations + 1):
            batch_input, batch_target = data
            self.call_plugins('batch', i, batch_input, batch_target)
            input_var = Variable(batch_input)
            target_var = Variable(batch_target)

            plugin_data = [None, None]

            def closure():
                batch_output = self.model(input_var)
                loss = self.criterion(batch_output, target_var)
                loss.backward()
                if plugin_data[0] is None:
                    plugin_data[0] = batch_output.data
                    plugin_data[1] = loss.data
                return loss

            self.optimizer.zero_grad()
            self.optimizer.step(closure)
            self.call_plugins('iteration', i, batch_input, batch_target,
                              *plugin_data)
            self.call_plugins('update', i, self.model)

        self.iterations += i



import math
import torch
from .Module import Module
from .Sequential import Sequential
from .SpatialZeroPadding import SpatialZeroPadding
from .SpatialConvolution import SpatialConvolution
from .SpatialConvolutionMap import SpatialConvolutionMap
from .Replicate import Replicate
from .CSubTable import CSubTable
from .CDivTable import CDivTable
from .utils import clear
import warnings


class SpatialSubtractiveNormalization(Module):

    def __init__(self, nInputPlane=1, kernel=None):
        super(SpatialSubtractiveNormalization, self).__init__()

        # get args
        self.nInputPlane = nInputPlane
        if kernel is None:
            kernel = torch.Tensor(9, 9).fill_(1)
        self.kernel = kernel
        kdim = self.kernel.ndimension()

        # check args
        if kdim != 2 and kdim != 1:
            raise ValueError('SpatialSubtractiveNormalization averaging kernel must be 2D or 1D')

        if (self.kernel.size(0) % 2) == 0 or (kdim == 2 and (self.kernel.size(1) % 2) == 0):
            raise ValueError('SpatialSubtractiveNormalization averaging kernel must have ODD dimensions')

        # normalize kernel
        self.kernel.div_(self.kernel.sum() * self.nInputPlane)

        # padding values
        padH = int(math.floor(self.kernel.size(0) / 2))
        padW = padH
        if kdim == 2:
            padW = int(math.floor(self.kernel.size(1) / 2))

        # create convolutional mean extractor
        self.meanestimator = Sequential()
        self.meanestimator.add(SpatialZeroPadding(padW, padW, padH, padH))
        if kdim == 2:
            self.meanestimator.add(SpatialConvolution(self.nInputPlane, 1, self.kernel.size(1), self.kernel.size(0)))
        else:
            # TODO: map
            self.meanestimator.add(SpatialConvolutionMap(
                SpatialConvolutionMap.maps.oneToOne(self.nInputPlane), self.kernel.size(0), 1))
            self.meanestimator.add(SpatialConvolution(self.nInputPlane, 1, 1, self.kernel.size(0)))

        self.meanestimator.add(Replicate(self.nInputPlane, 0))

        # set kernel and bias
        if kdim == 2:
            for i in range(self.nInputPlane):
                self.meanestimator.modules[1].weight[0][i] = self.kernel
            self.meanestimator.modules[1].bias.zero_()
        else:
            for i in range(self.nInputPlane):
                self.meanestimator.modules[1].weight[i].copy_(self.kernel, broadcast=False)
                self.meanestimator.modules[2].weight[0][i].copy_(self.kernel, broadcast=False)

            self.meanestimator.modules[1].bias.zero_()
            self.meanestimator.modules[2].bias.zero_()

        # other operation
        self.subtractor = CSubTable()
        self.divider = CDivTable()

        # coefficient array, to adjust side effects
        self.coef = torch.Tensor(1, 1, 1)

        self.ones = None
        self._coef = None

    def updateOutput(self, input):
        # compute side coefficients
        dim = input.dim()
        if (input.dim() + 1 != self.coef.dim() or
                (input.size(dim - 1) != self.coef.size(dim - 1)) or
                (input.size(dim - 2) != self.coef.size(dim - 2))):
            if self.ones is None:
                self.ones = input.new()
            if self._coef is None:
                self._coef = self.coef.new()

            self.ones.resize_as_(input[0:1]).fill_(1)
            coef = self.meanestimator.updateOutput(self.ones).squeeze(0)
            self._coef.resize_as_(coef).copy_(coef)  # make contiguous for view
            size = list(coef.size())
            size = [input.size(0)] + size
            self.coef = self._coef.view(1, *self._coef.size()).expand(*size)

        # compute mean
        self.localsums = self.meanestimator.updateOutput(input)
        self.adjustedsums = (self.divider.updateOutput(
            [self.localsums, self.coef.contiguous().view_as(self.localsums)]))
        self.output = self.subtractor.updateOutput([input, self.adjustedsums.contiguous().view_as(input)])

        return self.output

    def updateGradInput(self, input, gradOutput):
        # resize grad
        self.gradInput.resize_as_(input).zero_()

        # backprop through all modules
        gradsub = self.subtractor.updateGradInput([input, self.adjustedsums.contiguous().view_as(input)], gradOutput)
        graddiv = (self.divider.updateGradInput(
            [self.localsums, self.coef.contiguous().view_as(self.localsums)], gradsub[1]))
        size = self.meanestimator.updateGradInput(input, graddiv[0]).size()
        self.gradInput.add_(self.meanestimator.updateGradInput(input, graddiv[0]))
        self.gradInput.add_(gradsub[0])

        return self.gradInput

    def clearState(self):
        clear(self, 'ones', '_coef')
        self.meanestimator.clearState()
        return super(SpatialSubtractiveNormalization, self).clearState()



import torch
from .Module import Module
from .utils import clear


class VolumetricMaxPooling(Module):

    def __init__(self, kT, kW, kH, dT=None, dW=None, dH=None, padT=0, padW=0, padH=0):
        super(VolumetricMaxPooling, self).__init__()

        self.kT = kT
        self.kH = kH
        self.kW = kW
        self.dT = dT or kT
        self.dW = dW or kW
        self.dH = dH or kH

        self.padT = padT
        self.padW = padW
        self.padH = padH

        self.ceil_mode = False
        self.indices = torch.LongTensor()

    def ceil(self):
        self.ceil_mode = True
        return self

    def floor(self):
        self.ceil_mode = False
        return self

    def updateOutput(self, input):
        dims = input.dim()
        self.itime = input.size(dims - 3)
        self.iheight = input.size(dims - 2)
        self.iwidth = input.size(dims - 1)

        if self.indices is None:
            self.indices = input.new()
        self.indices = self.indices.long()
        self._backend.VolumetricMaxPooling_updateOutput(
            self._backend.library_state,
            input,
            self.output,
            self.indices,
            self.kT, self.kW, self.kH,
            self.dT, self.dW, self.dH,
            self.padT, self.padW, self.padH,
            self.ceil_mode
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        self._backend.VolumetricMaxPooling_updateGradInput(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradInput,
            self.indices,
            self.kT, self.kW, self.kH,
            self.dT, self.dW, self.dH,
            self.padT, self.padW, self.padH,
            self.ceil_mode
        )
        return self.gradInput

    def clearState(self):
        clear(self, 'indices')
        return super(VolumetricMaxPooling, self).clearState()

    def __repr__(self):
        s = super(VolumetricMaxPooling, self).__repr__()
        s += '({}x{}x{}, {}, {}, {}'.format(self.kT, self.kW, self.kH, self.dT, self.dW, self.dH)
        if self.padT != 0 or self.padW != 0 or self.padH != 0:
            s += ', {}, {}, {}'.format(self.padT, self.padW, self.padH)
        s += ')'
        return s



import torch
from .Module import Module
from .utils import clear, recursiveResizeAs


class MixtureTable(Module):

    def __init__(self, dim=1):
        super(MixtureTable, self).__init__()
        self.dim = dim
        self.size = torch.Size()
        self.size2 = torch.Size()
        self.batchSize = 0
        self.backwardSetup = False
        self.gradInput = []

        self._gaterView = None
        self._expert = None
        self._expertView = None
        self._sum = None
        self._expertView2 = None
        self._expert2 = None
        self.table = False

    def updateOutput(self, input):
        gaterInput, expertInputs = input

        # buffers
        if self._gaterView is None:
            self._gaterView = input[0].new()
        if self._expert is None:
            self._expert = input[0].new()
        if self._expertView is None:
            self._expertView = input[0].new()

        self.dimG = 1
        batchSize = gaterInput.size(0)

        if self.table or isinstance(expertInputs, list):
            self.table = True
            if gaterInput.size(self.dimG) != len(expertInputs):
                raise RuntimeError("Should be one gater output per expert")

            expertInput = expertInputs[0]
            if self.batchSize != batchSize:
                size = [1] * (expertInput.dim() + 1)
                if self.dimG > 0:
                    size[0] = gaterInput.size(0)
                size[self.dim] = gaterInput.size(self.dimG)
                self.size = torch.Size(size)
                self.output.resize_as_(expertInput)
                self.backwardSetup = False
                self.batchSize = batchSize

            self._gaterView = gaterInput.view(self.size)
            self.output.zero_()
            # multiply accumulate gater outputs by their commensurate expert
            for i, expertInput in enumerate(expertInputs):
                gate = self._gaterView.select(self.dim, i).expand_as(expertInput)
                self.output.addcmul_(expertInput, gate)
        else:
            if self.batchSize != batchSize:
                size = [1] * expertInputs.dim()
                if self.dimG > 0:
                    size[0] = gaterInput.size(0)
                size[self.dim] = gaterInput.size(self.dimG)
                self.size = torch.Size(size)
                self.output.resize_as_(expertInputs.select(self.dim, 0))
                self.batchSize = batchSize
                self.backwardSetup = False

            self._gaterView = gaterInput.view(self.size)
            torch.mul(self._gaterView.expand_as(expertInputs), expertInputs, out=self._expert)
            torch.sum(self._expert, self.dim, True, out=self.output)
            self.output.resize_as_(expertInputs.select(self.dim, 0))

        return self.output

    def updateGradInput(self, input, gradOutput):
        gaterInput, expertInputs = input
        recursiveResizeAs(self.gradInput, input)
        gaterGradInput, expertGradInputs = self.gradInput

        # buffers
        if self._sum is None:
            self._sum = input[0].new()
        if self._expertView2 is None:
            self._expertView2 = input[0].new()
        if self._expert2 is None:
            self._expert2 = input[0].new()

        if self.table:
            if not self.backwardSetup:
                for i, expertInput in enumerate(expertInputs):
                    expertGradInput = expertGradInputs[i] or expertInput.clone()
                    expertGradInput.resize_as_(expertInput)
                    expertGradInputs[i] = expertGradInput

                gaterGradInput.resize_as_(gaterInput)
                self.backwardSetup = True

            # like CMulTable, but with broadcasting
            for i, expertGradInput in enumerate(expertGradInputs):
                # gater updateGradInput
                torch.mul(gradOutput, expertInputs[i], out=self._expert)
                if self.dimG == 0:
                    self._expertView = self._expert.view(-1)
                else:
                    self._expertView = self._expert.view(gradOutput.size(0), -1)

                torch.sum(self._expertView, self.dimG, True, out=self._sum)
                if self.dimG == 0:
                    gaterGradInput[i] = self._sum.select(self.dimG, 0)
                else:
                    gaterGradInput.select(self.dimG, i).copy_(self._sum.select(self.dimG, 0))

                # expert updateGradInput
                gate = self._gaterView.select(self.dim, i).expand_as(expertGradInput)
                expertGradInput.mul_(gate, gradOutput)
        else:
            if not self.backwardSetup:
                size2 = list(expertInputs.size())
                size2[self.dim] = 1
                self.size2 = torch.Size(size2)
                gaterGradInput.resize_as_(gaterInput)
                self.backwardSetup = True

            # gater updateGradInput
            self._expertView = gradOutput.contiguous().view(torch.Size(self.size2))
            gradOutput = self._expertView.expand_as(expertInputs)
            torch.mul(gradOutput, expertInputs, out=self._expert)
            expert = self._expert.transpose(self.dim, self.dimG)
            if not expert.is_contiguous():
                self._expert2.resize_as_(expert)
                self._expert2.copy_(expert)
                expert = self._expert2
            if self.dimG == 0:
                self._expertView2 = expert.view(gaterInput.size(0), -1)
            else:
                self._expertView2 = expert.view(gaterInput.size(0), gaterInput.size(1), -1)

            torch.sum(self._expertView2, self.dimG + 1, True, out=gaterGradInput)
            gaterGradInput.resize_as_(gaterInput)

            # expert updateGradInput
            torch.mul(self._gaterView.expand_as(expertInputs), gradOutput, out=expertGradInputs)

        return self.gradInput

    def type(self, type, tensorCache=None):
        self._gaterView = None
        self._expert = None
        self._expertView = None
        self._sum = None
        self._expert2 = None
        self._expertView2 = None
        return super(MixtureTable, self).type(type, tensorCache)

    def clearState(self, ):
        clear(self, [
            '_gaterView',
            '_expert',
            '_expertView',
            '_sum',
            '_expert2',
            '_expertView2',
        ])
        return super(MixtureTable, self).clearState()



import torch
from .Module import Module
from .utils import clear


class Dropout(Module):

    def __init__(self, p=0.5, inplace=False):
        super(Dropout, self).__init__()
        self.p = p
        self.inplace = inplace
        self.train = True
        self.noise = torch.Tensor()

    def updateOutput(self, input):
        if self.inplace:
            self.output.set_(input)
        else:
            self.output.resize_as_(input).copy_(input)

        if self.p > 0 and self.train:
            self.noise.resize_as_(input)
            self.noise.bernoulli_(1 - self.p)
            self.noise.div_(1 - self.p)
            self.output.mul_(self.noise)

        return self.output

    def updateGradInput(self, input, gradOutput):
        if self.inplace:
            self.gradInput.set_(gradOutput)
        else:
            self.gradInput.resize_as_(gradOutput).copy_(gradOutput)

        if self.p > 0 and self.train:
            self.gradInput.mul_(self.noise)  # simply mask the gradients with the noise vector

        return self.gradInput

    def setp(self, p):
        self.p = p

    def __repr__(self):
        return super(Dropout, self).__repr__() + '({:.4f})'.format(self.p)

    def clearState(self):
        clear(self, 'noise')
        return super(Dropout, self).clearState()



import torch
from .Module import Module
from .utils import clear


class Sum(Module):

    def __init__(self, dimension=0, sizeAverage=False):
        super(Sum, self).__init__()
        self.dimension = dimension
        self.sizeAverage = sizeAverage
        self._gradOutput = None

    def _getPositiveDimension(self, input):
        dimension = self.dimension
        if dimension < 0:
            dimension = input.dim() + dimension
        return dimension

    def updateOutput(self, input):
        dimension = self._getPositiveDimension(input)

        torch.sum(input, dimension, out=self.output, keepdim=True)
        if self.sizeAverage:
            self.output.div_(input.size(dimension))
        if self.output.dim() > 1:
            self.output.set_(self.output.select(dimension, 0))

        return self.output

    def updateGradInput(self, input, gradOutput):
        dimension = self._getPositiveDimension(input)
        # zero-strides don't work with MKL/BLAS, so
        # don't set self.gradInput to zero-stride tensor.
        # Instead, do a deepcopy.
        size = list(input.size())
        size[dimension] = 1
        if not gradOutput.is_contiguous():
            if self._gradOutput is None:
                self._gradOutput = gradOutput.new()
            self._gradOutput.resize_as_(gradOutput).copy_(gradOutput)
            gradOutput = self._gradOutput

        gradOutput = gradOutput.view(*size)
        self.gradInput.resize_as_(input)
        self.gradInput.copy_(gradOutput.expand_as(input))
        if self.sizeAverage:
            self.gradInput.div_(input.size(dimension))

        return self.gradInput

    def clearState(self):
        clear(self, '_gradOutput')
        return super(Sum, self).clearState()



import torch
from .Module import Module


class Sqrt(Module):

    def __init__(self, b=0, eps=0):
        super(Sqrt, self).__init__()
        self.eps = b
        self.eps = eps

    def updateOutput(self, input):
        self._backend.Sqrt_updateOutput(
            self._backend.library_state,
            input,
            self.output,
            self.eps
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        self._backend.Sqrt_updateGradInput(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradInput,
            self.output
        )
        return self.gradInput



import torch
from torch.autograd.function import Function, InplaceFunction
from torch._thnn import type2backend
from torch.autograd.variable import Variable

from . import _all_functions


class PReLU(Function):

    @staticmethod
    def symbolic(g, input, weight):
        # TODO: Properly support numel in type()
        if all(s == 1 for s in weight.type().sizes()):
            raise RuntimeError("single weight shared among input channels not supported")
        return g.appendNode(g.create("PRelu", [input, weight]))

    @staticmethod
    def forward(ctx, input, weight):
        ctx._backend = type2backend[type(input)]
        output = input.new()
        ctx._backend.PReLU_updateOutput(
            ctx._backend.library_state,
            input,
            output,
            weight
        )
        ctx.save_for_backward(input, weight)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, weight = ctx.saved_variables
        # alternatively, we could recalculate _backend
        return PReLUBackward.apply(input, weight, grad_output, ctx._backend)


class PReLUBackward(Function):
    @staticmethod
    def forward(ctx, input, weight, grad_output, backend):
        ctx.save_for_backward(input, weight, grad_output)
        grad_input = input.new()
        backend.PReLU_updateGradInput(
            backend.library_state,
            input,
            grad_output,
            grad_input,
            weight
        )

        # TODO: this won't have to be zeroed in the future
        grad_weight = weight.new().resize_as_(weight).zero_()
        backend.PReLU_accGradParameters(
            backend.library_state,
            input,
            grad_output,
            grad_input,
            weight,
            grad_weight,
            1
        )
        return grad_input, grad_weight

    @staticmethod
    def backward(ctx, ggI, ggW):
        input, weight, gO = ctx.saved_variables
        positive_mask = (input > 0).type_as(ggI)
        nonpositive_mask = (input <= 0).type_as(ggW)
        # Explanation: Let input be i, weight be w, grad_output be gO.
        # f(i, w) = i  if i > 0
        #         = wi if i <= 0
        # df/di * gO  = gO      if i > 0      df/dw * g0 = 0      if i > 0
        #             = g0 * w  if i <= 0                = g0 * i  if i <= 0
        # The rest is taking derivatives of these wrt i, w, gO and summing/expanding properly.
        if weight.numel() == 1:
            # from PReLU.forward: num_parameters == 0 is used indicate that a
            # single weight is shared among all input channels.
            mask = positive_mask + nonpositive_mask * weight.expand_as(input)
            ggO = ggI * mask + ggW.expand_as(gO) * (nonpositive_mask * input)
            return ggW.expand_as(gO) * gO * nonpositive_mask, (ggI * gO * nonpositive_mask).sum(), ggO, None, None
        else:
            # Expand ggW to match size of ggI; a simple expand doesn't work because
            # ggW is the size of the input channel (dim==1 unless there is only 1 dimension).  For example,
            # let ggI be size (3,4,5,6,7) and ggW be size (4).  Then we unsqueeze ggW to be size (4,1,1,1)
            # so the expand succeeds.
            dims_to_unsqueeze = max(input.dim() - 2, 0)
            ggW_expanded = ggW
            for _ in range(dims_to_unsqueeze):
                ggW_expanded = ggW_expanded.unsqueeze(1)
            ggW_expanded = ggW_expanded.expand_as(ggI)

            gI = ggW_expanded * gO * nonpositive_mask

            gW = ggI * gO * nonpositive_mask
            if input.dim() > 1:
                gW = gW.sum(0)
            while gW.dim() > 1:
                gW = gW.sum(1)

            ggO = None
            if gO.requires_grad:
                # expand weight as input as in ggW/ggI above
                weight_expanded = weight
                for _ in range(dims_to_unsqueeze):
                    weight_expanded = weight_expanded.unsqueeze(1)
                weight_expanded = weight_expanded.expand_as(input)

                mask = positive_mask + nonpositive_mask * weight_expanded
                ggO = ggI * mask + ggW_expanded * nonpositive_mask * input
            return gI, gW, ggO, None, None


class RReLU(InplaceFunction):

    @staticmethod
    def forward(ctx, input, lower, upper, train, inplace):
        ctx.lower = lower
        ctx.upper = upper
        ctx.train = train
        ctx.inplace = inplace
        ctx._backend = type2backend[type(input)]
        if ctx.inplace:
            ctx.mark_dirty(input)
            output = input
        else:
            output = input.new(input.size())
        ctx.noise = input.new()
        ctx._backend.RReLU_updateOutput(
            ctx._backend.library_state,
            input,
            output,
            ctx.noise,
            ctx.lower,
            ctx.upper,
            ctx.train,
            ctx.inplace,
            torch.default_generator if not input.is_cuda else 0
        )
        ctx.save_for_backward(input)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_variables
        return (RReLUBackward.apply(input, grad_output, ctx.noise, ctx.lower, ctx.upper, ctx.train),
                None, None, None, None)


class RReLUBackward(Function):

    @staticmethod
    def forward(ctx, input, grad_output, noise, lower, upper, train):
        ctx.noise = noise
        ctx.lower = lower
        ctx.upper = upper
        ctx.train = train
        ctx._backend = type2backend[type(input)]
        ctx.save_for_backward(input)

        grad_input = input.new()
        ctx._backend.RReLU_updateGradInput(
            ctx._backend.library_state,
            input,
            grad_output,
            grad_input,
            ctx.noise,
            ctx.lower,
            ctx.upper,
            ctx.train,
            False
        )
        return grad_input

    @staticmethod
    def backward(ctx, ggI):
        input, = ctx.saved_variables

        gI = None

        positive_mask = (input > 0).type_as(ggI)
        nonpositive_mask = (input <= 0).type_as(ggI)
        mask = positive_mask + nonpositive_mask * Variable(ctx.noise)
        ggO = ggI * mask
        return gI, ggO, None, None, None, None


class SELU(InplaceFunction):
    alpha = 1.6732632423543772848170429916717
    scale = 1.0507009873554804934193349852946

    @staticmethod
    def forward(ctx, input, inplace):
        backend = type2backend[type(input)]
        if inplace:
            ctx.mark_dirty(input)
            output = input
        else:
            output = input.new(input.size())
        backend.ELU_updateOutput(
            backend.library_state,
            input,
            output,
            SELU.alpha,
            inplace,
        )
        output.mul_(SELU.scale)
        ctx.save_for_backward(input, output)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, output = ctx.saved_variables
        if grad_output.volatile:
            grad_input = Variable(input.data.new(input.size()), volatile=True)
            backend = type2backend[type(input.data)]
            backend.ELU_updateGradInput(
                backend.library_state,
                grad_output.data.mul(SELU.scale),
                grad_input.data,
                output.data.div(SELU.scale),
                SELU.alpha,
                False
            )
        else:
            positive_mask = (output > 0).type_as(grad_output)
            negative_mask = (output <= 0).type_as(grad_output)
            grad_input = grad_output * SELU.scale * (positive_mask +
                                                     negative_mask * (output / SELU.scale + SELU.alpha))
        return grad_input, None


_all_functions.append(PReLU)
_all_functions.append(PReLUBackward)
_all_functions.append(RReLU)
_all_functions.append(RReLUBackward)
_all_functions.append(SELU)



import torch
from torch.autograd.function import Function, InplaceFunction, once_differentiable
from torch._thnn import type2backend


class GRUFused(Function):
    @staticmethod
    def forward(ctx, input_gate, hidden_gate, hx, ibias=None, hbias=None):
        ctx.backend = type2backend[type(input_gate)]

        hy = input_gate.new()
        workspace = input_gate.new(hx.numel() * 5)

        ctx.has_bias = False
        if ibias is not None:
            ctx.has_bias = True
            if ibias.dim() == 1:
                ibias = ibias.unsqueeze(0)
            if hbias.dim() == 1:
                hbias = hbias.unsqueeze(0)

        ctx.backend.GRUFused_updateOutput(
            ctx.backend.library_state,
            input_gate, hidden_gate, ibias, hbias, hx, hy, workspace)

        ctx.workspace = workspace
        ctx.igate_size = input_gate.size()
        ctx.hgate_size = hidden_gate.size()

        return hy

    @staticmethod
    @once_differentiable
    def backward(ctx, gradOutput):
        ctx.backend = type2backend[type(gradOutput)]

        gradInputHx = gradOutput.new()
        gradInInput = gradOutput.new(*ctx.igate_size)
        gradInHidden = gradOutput.new(*ctx.hgate_size)

        ctx.backend.GRUFused_updateGradInput(
            ctx.backend.library_state,
            gradInInput, gradInHidden, gradOutput, gradInputHx, ctx.workspace)

        gb1 = gb2 = None
        if ctx.has_bias:
            gb1 = gradInInput.sum(0, keepdim=False)
            gb2 = gradInHidden.sum(0, keepdim=False)
        return gradInInput, gradInHidden, gradInputHx, gb1, gb2


class LSTMFused(Function):
    @staticmethod
    def forward(ctx, input_gate, hidden_gate, cx, ibias=None, hbias=None):
        ctx.backend = type2backend[type(input_gate)]
        hy = input_gate.new()
        cy = input_gate.new()

        ctx.has_bias = False
        if ibias is not None:
            ctx.has_bias = True
            if ibias.dim() == 1:
                ibias = ibias.unsqueeze(0)
            if hbias.dim() == 1:
                hbias = hbias.unsqueeze(0)

        # input_gate gets overwritten with some intermediate values to use in backwards
        ctx.backend.LSTMFused_updateOutput(
            ctx.backend.library_state,
            input_gate, hidden_gate,
            ibias, hbias,
            cx, hy, cy)

        ctx.hgate_size = hidden_gate.size()
        ctx.save_for_backward(input_gate, cx, cy)

        return hy, cy

    @staticmethod
    @once_differentiable
    def backward(ctx, *gradOutput):
        ctx.backend = type2backend[type(gradOutput[0])]
        gradInputCx = gradOutput[0].new()
        gradInGates = gradOutput[0].new(*ctx.hgate_size)

        saved_tens, cx, cy = ctx.saved_tensors
        ctx.backend.LSTMFused_updateGradInput(
            ctx.backend.library_state,
            saved_tens, gradInGates, cx, cy,
            gradOutput[0], gradOutput[1], gradInputCx)

        gb1 = gb2 = None
        if ctx.has_bias:
            gb1 = gradInGates.sum(0, keepdim=False)
            gb2 = gradInGates.sum(0, keepdim=False)

        return gradInGates, gradInGates, gradInputCx, gb1, gb2



import torch
from torch._thnn import type2backend
from torch.autograd import Function

from . import _all_functions
from .auto import _BCELoss
import warnings


def _resize_weight(ctx, target):
    ctx.old_weight = ctx.weight
    if ctx.weight is not None and target.dim() != 1:
        ctx.weight = ctx.weight.view(1, target.size(1)).expand_as(target)


def _unresize_weight(ctx):
    ctx.weight = ctx.old_weight
    del ctx.old_weight


# TODO: move this code to THNN and remove _BCELoss from auto.py
class BCELoss(_BCELoss):

    @staticmethod
    def forward(ctx, input, target, weight, *args):
        if not target.is_same_size(input):
            warnings.warn("Using a target size ({}) that is different to the input size ({}) is deprecated. "
                          "Please ensure they have the same size.".format(target.size(), input.size()))
        assert input.nelement() == target.nelement()
        ctx.weight = weight
        _resize_weight(ctx, target)
        result = _BCELoss.forward(ctx, input, target, ctx.weight, *args)
        _unresize_weight(ctx)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        target = ctx.saved_tensors[1]
        _resize_weight(ctx, target)
        result = _BCELoss.backward(ctx, grad_output)
        _unresize_weight(ctx)
        return result


_all_functions.append(BCELoss)



import torch
from .Module import Module


class CriterionTable(Module):

    def __init__(self, criterion):
        super(CriterionTable, self).__init__()
        self.criterion = criterion
        self.gradInput = [criterion.gradInput]

    def updateOutput(self, input):
        self.output = self.criterion.updateOutput(*input)
        return self.output

    def updateGradInput(self, input, grad_output):
        self.criterion.updateGradInput(*input)
        return self.gradInput



"""
Based on python-torchfile package.
https://github.com/bshillingford/python-torchfile

Copyright (c) 2016, Brendan Shillingford
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

TYPE_NIL = 0
TYPE_NUMBER = 1
TYPE_STRING = 2
TYPE_TABLE = 3
TYPE_TORCH = 4
TYPE_BOOLEAN = 5
TYPE_FUNCTION = 6
TYPE_RECUR_FUNCTION = 8
LEGACY_TYPE_RECUR_FUNCTION = 7


import struct
from array import array
from collections import namedtuple
from functools import wraps

import torch
import torch.legacy.nn as nn
import torch.cuda
from torch._thnn import type2backend
from torch._utils import _import_dotted_name

HAS_CUDA = torch.cuda.is_available()

LuaFunction = namedtuple('LuaFunction', ['size', 'dumped', 'upvalues'])


class hashable_uniq_dict(dict):
    """
    Subclass of dict with equality and hashing semantics changed:
    equality and hashing is purely by reference/instance, to match
    the behaviour of lua tables.

    Supports lua-style dot indexing.

    This way, dicts can be keys of other dicts.
    """

    def __hash__(self):
        return id(self)

    def __getattr__(self, key):
        return self.get(key)

    def __eq__(self, other):
        return id(self) == id(other)
    # TODO: dict's __lt__ etc. still exist


class TorchObject(object):
    """
    Simple torch object, used by `add_trivial_class_reader`.
    Supports both forms of lua-style indexing, i.e. getattr and getitem.
    Use the `torch_typename` method to get the object's torch class name.

    Equality is by reference, as usual for lua (and the default for Python
    objects).
    """

    def __init__(self, typename, obj):
        self._typename = typename
        self._obj = obj

    def __getattr__(self, k):
        return self._obj.get(k)

    def __getitem__(self, k):
        return self._obj.get(k)

    def torch_typename(self):
        return self._typename

    def __repr__(self):
        return "TorchObject(%s, %s)" % (self._typename, repr(self._obj))

    def __str__(self):
        return repr(self)

    def __dir__(self):
        keys = list(self._obj.keys())
        keys.append('torch_typename')
        return keys


reader_registry = {}


def get_python_class(typename):
    module, _, cls_name = typename.rpartition('.')
    if cls_name.startswith('Cuda'):
        module = module + '.cuda'
        cls_name = cls_name[4:]
        if cls_name == 'Storage' or cls_name == 'Tensor':
            cls_name = 'Float' + cls_name
    return _import_dotted_name(module + '.' + cls_name)


def make_tensor_reader(typename):
    python_class = get_python_class(typename)

    def read_tensor(reader, version):
        # source:
        # https://github.com/torch/torch7/blob/master/generic/Tensor.c#L1243
        ndim = reader.read_int()

        # read size:
        size = torch.LongStorage(reader.read_long_array(ndim))
        # read stride:
        stride = torch.LongStorage(reader.read_long_array(ndim))
        # storage offset:
        storage_offset = reader.read_long() - 1
        # read storage:
        storage = reader.read()

        if storage is None or ndim == 0 or len(size) == 0 or len(stride) == 0:
            # empty torch tensor
            return python_class()

        return python_class().set_(storage, storage_offset, torch.Size(size), tuple(stride))
    return read_tensor


def make_storage_reader(typename):
    python_class = get_python_class(typename)
    # TODO: be smarter about this
    element_size = python_class().element_size()

    def read_storage(reader, version):
        # source:
        # https://github.com/torch/torch7/blob/master/generic/Storage.c#L244
        size = reader.read_long() * element_size
        return python_class.from_buffer(reader.f.read(size), 'native')
    return read_storage


def register_torch_class(obj_kind, reader_factory):
    for t in ['Double', 'Float', 'Half', 'Long', 'Int', 'Short', 'Char', 'Byte']:
        for prefix in ['', 'Cuda']:
            if prefix == 'Cuda' and not HAS_CUDA:
                continue
            if t == 'Half' and prefix == '':
                continue
            if prefix == 'Cuda' and t == 'Float':
                cls_name = 'torch.Cuda' + obj_kind
            else:
                cls_name = 'torch.' + prefix + t + obj_kind
            reader_registry[cls_name] = reader_factory(cls_name)


register_torch_class('Storage', make_storage_reader)
register_torch_class('Tensor', make_tensor_reader)

################################################################################
# Reader function for tds.Vector and tds.Hash
################################################################################


def tds_Vec_reader(reader, version):
    length = reader.read_long()
    return [reader.read() for i in range(length)]


def tds_Hash_reader(reader, version):
    length = reader.read_long()
    obj = {}
    for i in range(length):
        k = reader.read()
        v = reader.read()
        obj[k] = v
    return obj


reader_registry['tds.Vec'] = tds_Vec_reader
reader_registry['tds.Hash'] = tds_Hash_reader

################################################################################
# Reader function for nn modules
################################################################################


def _load_backend(obj):
    if hasattr(obj, '_type'):
        obj._backend = type2backend[obj._type]
        return
    # Try to find tensor attributes and infer type from them
    for key in dir(obj):
        attr = getattr(obj, key)
        if torch.is_tensor(attr):
            try:
                obj._backend = type2backend[type(attr)]
            except KeyError:
                pass
    # Monkey patch the forward to capture the type of input
    updateOutput_orig = obj.updateOutput

    def updateOutput_patch(*args):
        input = args[0]
        while not torch.is_tensor(input):
            input = input[0]
        obj._backend = type2backend[type(input)]
        obj.updateOutput = updateOutput_orig
        return obj.updateOutput(*args)
    obj.updateOutput = updateOutput_patch


def nn_reader(cls):
    def read_nn_class(reader, version):
        obj = cls.__new__(cls)
        attributes = reader.read()
        obj.__dict__.update(attributes)
        _load_backend(obj)
        return obj
    return read_nn_class


reader_registry.update({('nn.' + name): nn_reader(module)
                        for name, module in nn.__dict__.items()
                        if name[0] != '_' and name[0].upper() == name[0]})


def custom_reader(cls):
    def reader_factory(fn):
        base = nn_reader(cls)

        def wrapper(reader, version):
            obj = base(reader, version)
            fn(reader, version, obj)
            return obj
        reader_registry['nn.' + cls.__name__] = wrapper
        return wrapper
    return reader_factory


def BatchNorm_reader(reader, version, obj):
    if version < 2 and hasattr(obj, 'running_std'):
        obj.running_var = obj.running_var.pow(-2).add(-obj.eps)
        del obj.running_std

for prefix in ['', 'Spatial', 'Volumetric']:
    name = prefix + 'BatchNormalization'
    custom_reader(getattr(nn, name))(BatchNorm_reader)


@custom_reader(nn.Transpose)
def Transpose_reader(reader, version, obj):
    obj.permutations = list(
        map(lambda swap: [swap[0] - 1, swap[1] - 1], obj.permutations))


@custom_reader(nn.SpatialDivisiveNormalization)
def SpatialDivisiveNormalization_reader(reader, version, obj):
    obj.stdestimator.modules[-2].dim += 1
    obj.meanestimator.modules[-1].dim += 1


@custom_reader(nn.SpatialContrastiveNormalization)
def SpatialContrastiveNormalization_reader(reader, version, obj):
    raise RuntimeError("loading of SpatialContrastiveNormalization is disabled for now")


@custom_reader(nn.GradientReversal)
def GradientReversal_reader(reader, version, obj):
    if version < 2:
        setattr(obj, 'lambda', 1)


@custom_reader(nn.VolumetricAveragePooling)
def VolumetricAveragePooling_reader(reader, version, obj):
    obj.padT, obj.padH, obj.padW = 0, 0, 0
    obj.ceil_mode = False
    obj.count_include_pad = True

################################################################################
# Functions for patching objects so that they work with legacy modules
################################################################################


def registry_addon(fn):
    def wrapper_factory(module_name, *args, **kwargs):
        module_name = 'nn.' + module_name
        build_fn = reader_registry[module_name]

        def wrapper(reader, version):
            obj = build_fn(reader, version)
            fn(obj, *args, **kwargs)
            return obj
        reader_registry[module_name] = wrapper
    return wrapper_factory


@registry_addon
def attr_map(obj, attribute_map):
    for src, dst in attribute_map.items():
        setattr(obj, dst, getattr(obj, src))
        delattr(obj, src)


@registry_addon
def ensure_attr(obj, *attrs):
    for attr in attrs:
        if not hasattr(obj, attr):
            setattr(obj, attr, None)


@registry_addon
def make_none_attr(obj, *attrs):
    for attr in attrs:
        setattr(obj, attr, None)


@registry_addon
def decrement(obj, *attrs):
    for attr in attrs:
        value = getattr(obj, attr)
        value -= 1
        setattr(obj, attr, value)


@registry_addon
def decrement_positive(obj, *attrs):
    for attr in attrs:
        value = getattr(obj, attr)
        if value > 0:
            value -= 1
        setattr(obj, attr, value)


@registry_addon
def storage_to_size(obj, *attrs):
    for attr in attrs:
        value = getattr(obj, attr)
        setattr(obj, attr, torch.Size(value))


@registry_addon
def ensure_type(obj, type_map):
    for attr, converter in type_map.items():
        value = getattr(obj, attr)
        setattr(obj, attr, getattr(value, converter)())


ensure_attr('Linear', 'bias', 'gradWeight', 'gradBias', 'addBuffer')
ensure_attr('CAddTable', 'inplace')
ensure_attr('SpatialFractionalMaxPooling', 'outW', 'outH', 'ratioW', 'ratioH')
ensure_attr('BatchNormalization', 'weight', 'bias', 'gradWeight', 'gradBias',
            'save_mean', 'save_std')
ensure_attr('SpatialBatchNormalization', 'weight', 'bias', 'gradWeight', 'gradBias',
            'save_mean', 'save_std')
ensure_attr('VolumetricBatchNormalization', 'weight', 'bias', 'gradWeight', 'gradBias')
ensure_attr('LookupTable', 'maxNorm', 'normType', '_gradOutput', '_sorted', '_indices')
ensure_attr('MixtureTable', 'table')
ensure_attr('WeightedEuclidean', 'fastBackward')
ensure_attr('VolumetricMaxPooling', 'ceil_mode')
ensure_attr('BCECriterion', 'buffer')
ensure_attr('SpatialClassNLLCriterion', 'weights')
ensure_attr('ClassNLLCriterion', 'weights')
ensure_attr('ParallelCriterion', 'repeatTarget')
ensure_attr('MultiMarginCriterion', 'weights')
ensure_attr('SpatialConvolution', 'bias', 'gradWeight', 'gradBias', '_gradOutput')
ensure_attr('SpatialCrossMapLRN', 'scale')
ensure_attr('Dropout', 'inplace')
make_none_attr('SpatialConvolution', 'finput', 'fgradInput', '_input')
attr_map('ReLU', {'val': 'value'})
attr_map('Threshold', {'val': 'value'})
attr_map('Unsqueeze', {'pos': 'dim'})
attr_map('HardShrink', {'lambda': 'lambd'})
attr_map('SoftShrink', {'lambda': 'lambd'})
attr_map('GradientReversal', {'lambda': 'lambd'})
attr_map('SpatialAdaptiveMaxPooling', {'H': 'h', 'W': 'w'})
decrement('Index', 'dimension')
decrement('SelectTable', 'index')
decrement('SplitTable', 'dimension')
decrement_positive('JoinTable', 'dimension')
decrement('Parallel', 'inputDimension', 'outputDimension')
decrement('Concat', 'dimension')
decrement('DepthConcat', 'dimension')
decrement('Squeeze', 'dim')
decrement('Unsqueeze', 'dim')
decrement('Replicate', 'dim')
decrement('MixtureTable', 'dim')
decrement('Narrow', 'dimension', 'index')
decrement('NarrowTable', 'offset')
decrement('LookupTable', 'paddingValue')
decrement('SpatialConvolutionMap', 'connTable')
decrement('SpatialFullConvolutionMap', 'connTable')
decrement('Select', 'dimension', 'index')
decrement('Padding', 'dim', 'index')
decrement('PartialLinear', 'partition')
decrement_positive('Sum', 'dimension')
decrement_positive('Max', 'dimension')
decrement_positive('Min', 'dimension')
decrement_positive('Mean', 'dimension')
storage_to_size('View', 'size')
storage_to_size('DepthConcat', 'outputSize')
storage_to_size('MixtureTable', 'size')
ensure_type('PartialLinear', {'partition': 'long'})


class T7ReaderException(Exception):
    pass


class T7Reader:

    def __init__(self,
                 fileobj,
                 list_heuristic=True,
                 int_heuristic=True,
                 unknown_classes=False,
                 long_size=None):
        """
        Params:
        * `fileobj` file object to read from, must be actual file object
                    as it must support array, struct, and numpy
        * `list_heuristic`: automatically turn tables with only consecutive
                                positive integral indices into lists
                                (default True)
        * `int_heuristic`: cast all whole floats into ints (default True)
        * `force_deserialize_classes`: deserialize all classes, not just the
                                       whitelisted ones (default True)
        """
        self.f = fileobj
        self.memo = {}

        self.list_heuristic = list_heuristic
        self.int_heuristic = int_heuristic
        self.unknown_classes = unknown_classes
        self.long_size = long_size

    def _read(self, fmt):
        sz = struct.calcsize(fmt)
        result = struct.unpack(fmt, self.f.read(sz))
        if len(result) == 1:
            return result[0]
        return result

    def read_boolean(self):
        return self.read_int() == 1

    def read_int(self):
        return self._read('i')

    def read_long(self):
        if self.long_size is None:
            return self._read('l')
        elif self.long_size is 8:
            return self._read('q')
        else:
            return self._read('i')

    def read_long_array(self, n):
        if self.long_size is not None:
            lst = []
            for i in range(n):
                lst.append(self.read_long())
            return lst
        else:
            arr = array('l')
            arr.fromfile(self.f, n)
            return arr.tolist()

    def read_float(self):
        return self._read('f')

    def read_double(self):
        return self._read('d')

    def read_string(self):
        size = self.read_int()
        byte_str = self.f.read(size)
        if not isinstance(byte_str, str):
            byte_str = str(byte_str, 'ascii')
        return byte_str

    def read_number(self):
        x = self.read_double()
        # Extra checking for integral numbers:
        if self.int_heuristic and x.is_integer():
            return int(x)
        return x

    def memoize_index(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            index = self.read_int()
            if index in self.memo:
                return self.memo[index]
            result = fn(self, *args, **kwargs)
            self.memo[index] = result
            return result
        return wrapper

    @memoize_index
    def read_function(self):
        size = self.read_int()
        dumped = self.f.read(size)
        upvalues = self.read()
        return LuaFunction(size, dumped, upvalues)

    @memoize_index
    def read_object(self):
        version_str = self.read_string()
        if version_str.startswith('V '):
            version = int(version_str.partition(' ')[2])
            cls_name = self.read_string()
        else:
            cls_name = version_str
            version = 0  # created before existence of versioning

        if cls_name in reader_registry:
            return reader_registry[cls_name](self, version)
        if self.unknown_classes:
            return TorchObject(cls_name, self.read())
        raise T7ReaderException(("don't know how to deserialize Lua class "
                                 "{}. If you want to ignore this error and load this object "
                                 "as a dict, specify unknown_classes=True in reader's "
                                 "constructor").format(cls_name))

    def _can_be_list(self, table):
        def is_natural(key):
            return (isinstance(key, int) or
                    (isinstance(key, float) and key.is_integer()) and
                    k > 0)
        natural_keys = all(map(is_natural, table.keys()))
        if not natural_keys:
            return False
        key_sum = sum(table.keys())
        n = len(table)
        return n * (n + 1) == 2 * key_sum

    @memoize_index
    def read_table(self):
        size = self.read_int()
        table = hashable_uniq_dict()  # custom hashable dict, can be a key
        for i in range(size):
            k = self.read()
            v = self.read()
            table[k] = v
        if self.list_heuristic and self._can_be_list(table):
            return [table[i] for i in range(1, len(table) + 1)]
        return table

    def read(self):
        typeidx = self.read_int()

        if typeidx == TYPE_NIL:
            return None
        elif typeidx == TYPE_NUMBER:
            return self.read_number()
        elif typeidx == TYPE_BOOLEAN:
            return self.read_boolean()
        elif typeidx == TYPE_STRING:
            return self.read_string()
        elif (typeidx == TYPE_FUNCTION or typeidx == TYPE_RECUR_FUNCTION or
              typeidx == LEGACY_TYPE_RECUR_FUNCTION):
            return self.read_function()
        elif typeidx == TYPE_TORCH:
            return self.read_object()
        elif typeidx == TYPE_TABLE:
            return self.read_table()
        else:
            raise T7ReaderException("unknown type id {}. The file may be "
                                    "corrupted.".format(typeidx))


def load_lua(filename, **kwargs):
    """
    Loads the given t7 file using default settings; kwargs are forwarded
    to `T7Reader`.
    """
    with open(filename, 'rb') as f:
        reader = T7Reader(f, **kwargs)
        return reader.read()



import math
import torch
from .Module import Module


class Add(Module):

    def __init__(self, inputSize, scalar=False):
        super(Add, self).__init__()
        size = inputSize
        if scalar:
            assert size == 1
        self.scalar = scalar
        self.bias = torch.Tensor(size)
        self.gradBias = torch.Tensor(size)

        self._ones = torch.Tensor((1,))

        self.reset()

    def reset(self, stdv=None):
        if stdv is not None:
            stdv = stdv * math.sqrt(3)
        else:
            stdv = 1. / math.sqrt(self.bias.size(0))

        self.bias.uniform_(-stdv, stdv)

    def updateOutput(self, input):
        self.output.resize_as_(input).copy_(input)
        if self.scalar:
            self.output.add_(self.bias[0])
        else:
            batchSize = input.size(0)
            if self._ones.size(0) != batchSize:
                self._ones.resize_(batchSize).fill_(1)

            bias = self.bias.view(-1)
            output = self.output.view(batchSize, -1)
            output.addr_(self._ones, bias)

        return self.output

    def updateGradInput(self, input, gradOutput):
        if self.gradInput is not None:
            self.gradInput.resize_as_(gradOutput).copy_(gradOutput)
            return self.gradInput

    def accGradParameters(self, input, gradOutput, scale=1):
        if self.gradBias.size(0) == 1:
            self.gradBias[0] = self.gradBias[0] + scale * gradOutput.sum()
        else:
            if input.is_same_size(self.bias):
                self.gradBias.add_(scale, gradOutput)
            else:
                gradOutput = gradOutput.contiguous().view(input.size(0), -1)
                self.gradBias.view(-1).addmv_(scale, gradOutput.t(), self._ones)



import torch
from .Module import Module


class CAddTable(Module):

    def __init__(self, inplace=False):
        super(CAddTable, self).__init__()
        self.inplace = inplace
        self.gradInput = []

    def updateOutput(self, input):
        if self.inplace:
            self.output.set_(input[0])
        else:
            self.output.resize_as_(input[0]).copy_(input[0])

        for i in range(1, len(input)):
            self.output.add_(input[i])

        return self.output

    def updateGradInput(self, input, gradOutput):
        for i in range(len(input)):
            if i >= len(self.gradInput):
                assert i == len(self.gradInput)
                self.gradInput.append(input[0].new())

            if self.inplace:
                self.gradInput[i].set_(gradOutput)
            else:
                self.gradInput[i].resize_as_(input[i]).copy_(gradOutput)

        del self.gradInput[len(input):]

        return self.gradInput




class Plugin(object):

    def __init__(self, interval=None):
        if interval is None:
            interval = []
        self.trigger_interval = interval

    def register(self, trainer):
        raise NotImplementedError



import warnings
import torch
from torch.nn.parameter import Parameter

from .module import Module
from .. import functional as F


class Threshold(Module):
    r"""Thresholds each element of the input Tensor

    Threshold is defined as::

         y =  x        if x >  threshold
              value    if x <= threshold

    Args:
        threshold: The value to threshold at
        value: The value to replace with
        inplace: can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Threshold(0.1, 20)
        >>> input = Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, threshold, value, inplace=False):
        super(Threshold, self).__init__()
        self.threshold = threshold
        self.value = value
        self.inplace = inplace
        # TODO: check in THNN (if inplace == True, then assert value <= threshold)

    def forward(self, input):
        return F.threshold(input, self.threshold, self.value, self.inplace)

    def __repr__(self):
        inplace_str = ', inplace' if self.inplace else ''
        return self.__class__.__name__ + ' (' \
            + str(self.threshold) \
            + ', ' + str(self.value) \
            + inplace_str + ')'


class ReLU(Threshold):
    r"""Applies the rectified linear unit function element-wise
    :math:`{ReLU}(x)= max(0, x)`

    Args:
        inplace: can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.ReLU()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, inplace=False):
        super(ReLU, self).__init__(0, 0, inplace)

    def __repr__(self):
        inplace_str = 'inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + inplace_str + ')'


class RReLU(Module):

    def __init__(self, lower=1. / 8, upper=1. / 3, inplace=False):
        super(RReLU, self).__init__()
        self.lower = lower
        self.upper = upper
        self.inplace = inplace

    def forward(self, input):
        return F.rrelu(input, self.lower, self.upper, self.training, self.inplace)

    def __repr__(self):
        inplace_str = ', inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + str(self.lower) \
            + ', ' + str(self.upper) \
            + inplace_str + ')'


class Hardtanh(Module):
    r"""Applies the HardTanh function element-wise

    HardTanh is defined as::

       f(x) = +1, if x  >  1
       f(x) = -1, if x  < -1
       f(x) =  x,  otherwise

    The range of the linear region :math:`[-1, 1]` can be adjusted

    Args:
        min_val: minimum value of the linear region range. Default: -1
        max_val: maximum value of the linear region range. Default: 1
        inplace: can optionally do the operation in-place. Default: ``False``

    Keyword arguments :attr:`min_value` and :attr:`max_value`
    have been deprecated in favor of :attr:`min_val` and :attr:`max_val`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Hardtanh(-2, 2)
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, min_val=-1, max_val=1, inplace=False, min_value=None, max_value=None):
        super(Hardtanh, self).__init__()
        if min_value is not None:
            warnings.warn("keyword argument min_value is deprecated and renamed to min_val")
            min_val = min_value
        if max_value is not None:
            warnings.warn("keyword argument max_value is deprecated and renamed to max_val")
            max_val = max_value

        self.min_val = min_val
        self.max_val = max_val
        self.inplace = inplace
        assert self.max_val > self.min_val

    def forward(self, input):
        return F.hardtanh(input, self.min_val, self.max_val, self.inplace)

    def __repr__(self):
        inplace_str = ', inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + 'min_val=' + str(self.min_val) \
            + ', max_val=' + str(self.max_val) \
            + inplace_str + ')'


class ReLU6(Hardtanh):
    r"""Applies the element-wise function :math:`{ReLU6}(x) = min(max(0,x), 6)`

    Args:
        inplace: can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.ReLU6()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, inplace=False):
        super(ReLU6, self).__init__(0, 6, inplace)

    def __repr__(self):
        inplace_str = 'inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + inplace_str + ')'


class Sigmoid(Module):
    r"""Applies the element-wise function :math:`f(x) = 1 / ( 1 + exp(-x))`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Sigmoid()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        return torch.sigmoid(input)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Tanh(Module):
    r"""Applies element-wise,
    :math:`f(x) = (exp(x) - exp(-x)) / (exp(x) + exp(-x))`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Tanh()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        return torch.tanh(input)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class ELU(Module):
    r"""Applies element-wise,
    :math:`f(x) = max(0,x) + min(0, alpha * (exp(x) - 1))`

    Args:
        alpha: the alpha value for the ELU formulation. Default: 1.0
        inplace: can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.ELU()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, alpha=1., inplace=False):
        super(ELU, self).__init__()
        self.alpha = alpha
        self.inplace = inplace

    def forward(self, input):
        return F.elu(input, self.alpha, self.inplace)

    def __repr__(self):
        inplace_str = ', inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + 'alpha=' + str(self.alpha) \
            + inplace_str + ')'


class SELU(Module):
    r"""Applies element-wise,
    :math:`f(x) = scale * (\max(0,x) + \min(0, alpha * (\exp(x) - 1)))`,
    with ``alpha=1.6732632423543772848170429916717`` and
    ``scale=1.0507009873554804934193349852946``.

    More details can be found in the paper `Self-Normalizing Neural Networks`_ .

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.SELU()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))

    .. _Self-Normalizing Neural Networks: https://arxiv.org/abs/1706.02515
    """

    def __init__(self, inplace=False):
        super(SELU, self).__init__()
        self.inplace = inplace

    def forward(self, input):
        return F.selu(input, self.inplace)

    def __repr__(self):
        inplace_str = '(inplace)' if self.inplace else ''
        return self.__class__.__name__ + inplace_str


class GLU(Module):
    r"""Applies the gated linear unit function
    :math:`{GLU}(a, b)= a \otimes \sigma(b)` where `a` is the first half of
    the input vector and `b` is the second half.

    Args:
        dim (int): the dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(*, N / 2, *)`

    Examples::

        >>> m = nn.GLU()
        >>> input = autograd.Variable(torch.randn(4, 2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, dim=-1):
        super(GLU, self).__init__()
        self.dim = dim

    def forward(self, input):
        return F.glu(input, self.dim)

    def __repr__(self):
        return '{}(dim={})'.format(self.__class__.__name__, self.dim)


class Hardshrink(Module):
    r"""Applies the hard shrinkage function element-wise
    Hardshrink is defined as::
        f(x) = x, if x >  lambda
        f(x) = x, if x < -lambda
        f(x) = 0, otherwise

    Args:
        lambd: the lambda value for the Hardshrink formulation. Default: 0.5

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Hardshrink()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, lambd=0.5):
        super(Hardshrink, self).__init__()
        self.lambd = lambd

    def forward(self, input):
        return F.hardshrink(input, self.lambd)

    def __repr__(self):
        return self.__class__.__name__ + '(' \
            + str(self.lambd) + ')'


class LeakyReLU(Module):
    r"""Applies element-wise,
    :math:`f(x) = max(0, x) + {negative\_slope} * min(0, x)`

    Args:
        negative_slope: Controls the angle of the negative slope. Default: 1e-2
        inplace: can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.LeakyReLU(0.1)
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, negative_slope=1e-2, inplace=False):
        super(LeakyReLU, self).__init__()
        self.negative_slope = negative_slope
        self.inplace = inplace

    def forward(self, input):
        return F.leaky_relu(input, self.negative_slope, self.inplace)

    def __repr__(self):
        inplace_str = ', inplace' if self.inplace else ''
        return self.__class__.__name__ + '(' \
            + str(self.negative_slope) \
            + inplace_str + ')'


class LogSigmoid(Module):
    r"""Applies element-wise :math:`LogSigmoid(x) = log( 1 / (1 + exp(-x_i)))`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.LogSigmoid()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        return F.logsigmoid(input)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Softplus(Module):
    r"""Applies element-wise :math:`f(x) = 1/beta * log(1 + exp(beta * x_i))`

    SoftPlus is a smooth approximation to the ReLU function and can be used
    to constrain the output of a machine to always be positive.

    For numerical stability the implementation reverts to the linear function
    for inputs above a certain value.

    Args:
        beta: the beta value for the Softplus formulation. Default: 1
        threshold: values above this revert to a linear function. Default: 20

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Softplus()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, beta=1, threshold=20):
        super(Softplus, self).__init__()
        self.beta = beta
        self.threshold = threshold

    def forward(self, input):
        return F.softplus(input, self.beta, self.threshold)

    def __repr__(self):
        return self.__class__.__name__ + '(' \
            + 'beta=' + str(self.beta) \
            + ', threshold=' + str(self.threshold) + ')'


class Softshrink(Module):
    r"""Applies the soft shrinkage function elementwise

    SoftShrinkage operator is defined as::

        f(x) = x-lambda, if x > lambda >  f(x) = x+lambda, if x < -lambda
        f(x) = 0, otherwise

    Args:
        lambd: the lambda value for the Softshrink formulation. Default: 0.5

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Softshrink()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, lambd=0.5):
        super(Softshrink, self).__init__()
        self.lambd = lambd

    def forward(self, input):
        return F.softshrink(input, self.lambd)

    def __repr__(self):
        return self.__class__.__name__ + '(' \
            + str(self.lambd) + ')'


class PReLU(Module):
    r"""Applies element-wise the function
    :math:`PReLU(x) = max(0,x) + a * min(0,x)` Here "a" is a learnable
    parameter. When called without arguments, nn.PReLU() uses a single
    parameter "a" across all input channels. If called with nn.PReLU(nChannels),
    a separate "a" is used for each input channel.


    .. note::
        weight decay should not be used when learning "a" for good performance.

    Args:
        num_parameters: number of "a" to learn. Default: 1
        init: the initial value of "a". Default: 0.25

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.PReLU()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, num_parameters=1, init=0.25):
        self.num_parameters = num_parameters
        super(PReLU, self).__init__()
        self.weight = Parameter(torch.Tensor(num_parameters).fill_(init))

    def forward(self, input):
        return F.prelu(input, self.weight)

    def __repr__(self):
        return self.__class__.__name__ + '(' \
            + 'num_parameters=' + str(self.num_parameters) + ')'


class Softsign(Module):
    r"""Applies element-wise, the function :math:`f(x) = x / (1 + |x|)`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Softsign()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        return F.softsign(input)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Tanhshrink(Module):
    r"""Applies element-wise, :math:`Tanhshrink(x) = x - Tanh(x)`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    Examples::

        >>> m = nn.Tanhshrink()
        >>> input = autograd.Variable(torch.randn(2))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        return F.tanhshrink(input)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Softmin(Module):
    r"""Applies the Softmin function to an n-dimensional input Tensor
    rescaling them so that the elements of the n-dimensional output Tensor
    lie in the range `(0, 1)` and sum to 1

    :math:`f(x) = \frac{\exp(-x_i)}{\sum_j \exp(-x_j)}`

    Shape:
        - Input: any shape
        - Output: same as input

    Arguments:
        dim (int): A dimension along which Softmax will be computed (so every slice
            along dim will sum to 1).

    Returns:
        a Tensor of the same dimension and shape as the input, with
        values in the range [0, 1]

    Examples::

        >>> m = nn.Softmin()
        >>> input = autograd.Variable(torch.randn(2, 3))
        >>> print(input)
        >>> print(m(input))
    """
    def __init__(self, dim=None):
        super(Softmin, self).__init__()
        self.dim = dim

    def forward(self, input):
        return F.softmin(input, self.dim, _stacklevel=5)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Softmax(Module):
    r"""Applies the Softmax function to an n-dimensional input Tensor
    rescaling them so that the elements of the n-dimensional output Tensor
    lie in the range (0,1) and sum to 1

    Softmax is defined as
    :math:`f_i(x) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}`

    Shape:
        - Input: any shape
        - Output: same as input

    Returns:
        a Tensor of the same dimension and shape as the input with
        values in the range [0, 1]

    Arguments:
        dim (int): A dimension along which Softmax will be computed (so every slice
            along dim will sum to 1).

    .. note::
        This module doesn't work directly with NLLLoss,
        which expects the Log to be computed between the Softmax and itself.
        Use Logsoftmax instead (it's faster and has better numerical properties).

    Examples::

        >>> m = nn.Softmax()
        >>> input = autograd.Variable(torch.randn(2, 3))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, dim=None):
        super(Softmax, self).__init__()
        self.dim = dim

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not hasattr(self, 'dim'):
            self.dim = None

    def forward(self, input):
        return F.softmax(input, self.dim, _stacklevel=5)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Softmax2d(Module):
    r"""Applies SoftMax over features to each spatial location

    When given an image of Channels x Height x Width, it will

    apply Softmax to each location :math:`(Channels, h_i, w_j)`

    Shape:
        - Input: :math:`(N, C, H, W)`
        - Output: :math:`(N, C, H, W)` (same shape as input)

    Returns:
        a Tensor of the same dimension and shape as the input with
        values in the range [0, 1]

    Examples::

        >>> m = nn.Softmax2d()
        >>> # you softmax over the 2nd dimension
        >>> input = autograd.Variable(torch.randn(2, 3, 12, 13))
        >>> print(input)
        >>> print(m(input))
    """

    def forward(self, input):
        assert input.dim() == 4, 'Softmax2d requires a 4D tensor as input'
        return F.softmax(input, 1, _stacklevel=5)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class LogSoftmax(Module):
    r"""Applies the Log(Softmax(x)) function to an n-dimensional input Tensor.
    The LogSoftmax formulation can be simplified as

    :math:`f_i(x) = log(exp(x_i) / sum_j exp(x_j) )`

    Shape:
        - Input: any shape
        - Output: same as input

    Arguments:
        dim (int): A dimension along which Softmax will be computed (so every slice
            along dim will sum to 1).

    Returns:
        a Tensor of the same dimension and shape as the input with
        values in the range [-inf, 0)

    Examples::

        >>> m = nn.LogSoftmax()
        >>> input = autograd.Variable(torch.randn(2, 3))
        >>> print(input)
        >>> print(m(input))
    """

    def __init__(self, dim=None):
        super(LogSoftmax, self).__init__()
        self.dim = dim

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not hasattr(self, 'dim'):
            self.dim = None

    def forward(self, input):
        return F.log_softmax(input, self.dim, _stacklevel=5)

    def __repr__(self):
        return self.__class__.__name__ + '()'



import torch
from .Criterion import Criterion


class CosineEmbeddingCriterion(Criterion):

    def __init__(self, margin=0, sizeAverage=True):
        super(CosineEmbeddingCriterion, self).__init__()
        self.margin = margin
        self.sizeAverage = sizeAverage
        self.gradInput = [torch.Tensor(), torch.Tensor()]
        self.buffer = None
        self.w1 = None
        self.w22 = None
        self.w = None
        self.w32 = None
        self._outputs = None
        self._idx = None

    def updateOutput(self, input, y):
        input1, input2 = input[0], input[1]

        # keep backward compatibility
        if self.buffer is None:
            self.buffer = input1.new()
            self.w1 = input1.new()
            self.w22 = input1.new()
            self.w = input1.new()
            self.w32 = input1.new()
            self._outputs = input1.new()

            # comparison operators behave differently from cuda/c implementations
            # TODO: verify name
            if input1.type() == 'torch.cuda.FloatTensor':
                self._idx = torch.cuda.ByteTensor()
            else:
                self._idx = torch.ByteTensor()

        torch.mul(input1, input2, out=self.buffer)
        torch.sum(self.buffer, 1, out=self.w1, keepdim=True)

        epsilon = 1e-12
        torch.mul(input1, input1, out=self.buffer)
        torch.sum(self.buffer, 1, out=self.w22, keepdim=True).add_(epsilon)
        # self._outputs is also used as a temporary buffer
        self._outputs.resize_as_(self.w22).fill_(1)
        torch.div(self._outputs, self.w22, out=self.w22)
        self.w.resize_as_(self.w22).copy_(self.w22)

        torch.mul(input2, input2, out=self.buffer)
        torch.sum(self.buffer, 1, out=self.w32, keepdim=True).add_(epsilon)
        torch.div(self._outputs, self.w32, out=self.w32)
        self.w.mul_(self.w32)
        self.w.sqrt_()

        torch.mul(self.w1, self.w, out=self._outputs)
        self._outputs = self._outputs.select(1, 0)

        torch.eq(y, -1, out=self._idx)
        self._outputs[self._idx] = self._outputs[self._idx].add_(-self.margin).clamp_(min=0)
        torch.eq(y, 1, out=self._idx)
        self._outputs[self._idx] = self._outputs[self._idx].mul_(-1).add_(1)

        self.output = self._outputs.sum()

        if self.sizeAverage:
            self.output = self.output / y.size(0)

        return self.output

    def updateGradInput(self, input, y):
        v1 = input[0]
        v2 = input[1]

        gw1 = self.gradInput[0]
        gw2 = self.gradInput[1]
        gw1.resize_as_(v1).copy_(v2)
        gw2.resize_as_(v1).copy_(v1)

        torch.mul(self.w1, self.w22, out=self.buffer)
        gw1.addcmul_(-1, self.buffer.expand_as(v1), v1)
        gw1.mul_(self.w.expand_as(v1))

        torch.mul(self.w1, self.w32, out=self.buffer)
        gw2.addcmul_(-1, self.buffer.expand_as(v1), v2)
        gw2.mul_(self.w.expand_as(v1))

        # self._idx = self._outputs <= 0
        torch.le(self._outputs, 0, out=self._idx)
        self._idx = self._idx.view(-1, 1).expand(gw1.size())
        gw1[self._idx] = 0
        gw2[self._idx] = 0

        torch.eq(y, 1, out=self._idx)
        self._idx = self._idx.view(-1, 1).expand(gw2.size())
        gw1[self._idx] = gw1[self._idx].mul_(-1)
        gw2[self._idx] = gw2[self._idx].mul_(-1)

        if self.sizeAverage:
            gw1.div_(y.size(0))
            gw2.div_(y.size(0))

        return self.gradInput

    def type(self, type=None, tensorCache=None):
        if not type:
            return self._type

        self._idx = None
        super(CosineEmbeddingCriterion, self).type(type, tensorCache)
        # comparison operators behave differently from cuda/c implementations
        if type == 'torch.cuda.FloatTensor':
            self._idx = torch.cuda.ByteTensor()
        else:
            self._idx = torch.ByteTensor()

        return self



import math
import torch
from .Module import Module


class TemporalConvolution(Module):

    def __init__(self, inputFrameSize, outputFrameSize, kW, dW=1):
        super(TemporalConvolution, self).__init__()

        self.inputFrameSize = inputFrameSize
        self.outputFrameSize = outputFrameSize
        self.kW = kW
        self.dW = dW

        self.weight = torch.Tensor(outputFrameSize, inputFrameSize * kW)
        self.bias = torch.Tensor(outputFrameSize)
        self.gradWeight = torch.Tensor(outputFrameSize, inputFrameSize * kW)
        self.gradBias = torch.Tensor(outputFrameSize)

        self.reset()

    def reset(self, stdv=None):
        if stdv is not None:
            stdv = stdv * math.sqrt(3)
        else:
            stdv = 1. / math.sqrt(self.kW * self.inputFrameSize)

        self.weight.uniform_(-stdv, stdv)
        self.bias.uniform_(-stdv, stdv)

    def updateOutput(self, input):
        self._backend.TemporalConvolution_updateOutput(
            self._backend.library_state,
            input,
            self.output,
            self.weight,
            self.bias,
            self.kW,
            self.dW,
            self.inputFrameSize,
            self.outputFrameSize
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        if self.gradInput is None:
            return
        self._backend.TemporalConvolution_updateGradInput(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradInput,
            self.weight,
            self.kW,
            self.dW
        )
        return self.gradInput

    def accGradParameters(self, input, gradOutput, scale=1):
        self._backend.TemporalConvolution_accGradParameters(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradWeight,
            self.gradBias,
            self.kW,
            self.dW,
            scale
        )



import torch
from .SpatialConvolution import SpatialConvolution


class SpatialDilatedConvolution(SpatialConvolution):

    def __init__(self, nInputPlane, nOutputPlane, kW, kH, dW=1, dH=1, padW=0, padH=None, dilationH=1, dilationW=None):
        super(SpatialDilatedConvolution, self).__init__(nInputPlane, nOutputPlane, kW, kH, dW, dH, padW, padH)

        self.dilationH = dilationH
        self.dilationW = dilationW if dilationW is not None else dilationH

    def updateOutput(self, input):
        if self.finput is None:
            self.finput = self.weight.new()
        if self.fgradInput is None:
            self.fgradInput = self.weight.new()
        input = self._makeContiguous(input)
        self._backend.SpatialDilatedConvolution_updateOutput(
            self._backend.library_state,
            input,
            self.output,
            self.weight,
            self.bias,
            self.finput,
            self.fgradInput,
            self.kW, self.kH,
            self.dW, self.dH,
            self.padW, self.padH,
            self.dilationH, self.dilationW
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        if self.gradInput is None:
            return

        input, gradOutput = self._makeContiguous(input, gradOutput)
        if self.fgradInput is None:
            self.fgradInput = self.weight.new()
        self._backend.SpatialDilatedConvolution_updateGradInput(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradInput,
            self.weight,
            self.finput,
            self.kW, self.kH,
            self.dW, self.dH,
            self.padW, self.padH,
            self.dilationH, self.dilationW
        )
        return self.gradInput

    def accGradParameters(self, input, gradOutput, scale=1):
        input, gradOutput = self._makeContiguous(input, gradOutput)
        if self.fgradInput is None:
            self.fgradInput = self.weight.new()
        self._backend.SpatialDilatedConvolution_accGradParameters(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradWeight,
            self.gradBias,
            self.finput,
            self.fgradInput,
            self.kW, self.kH,
            self.dW, self.dH,
            self.padW, self.padH,
            self.dilationH, self.dilationW,
            scale
        )

    def __repr__(self):
        s = super(SpatialConvolution, self).__repr__()
        s += '({} -> {}, {}x{}'.format(self.nInputPlane, self.nOutputPlane, self.kW, self.kH)
        if self.dW != 1 or self.dH != 1 or self.padW != 0 or self.padH != 0:
            s += ', {}, {}'.format(self.dW, self.dH)

        if self.padW != 0 or self.padH != 0:
            s += ', {}, {}'.format(self.padW, self.padH)

        s += ', {}, {}'.format(self.dilationW, self.dilationH)

        s += ')'
        if self.bias is None:
            s += ' without bias'
        return s



import torch
from .Module import Module


class CDivTable(Module):

    def __init__(self, ):
        super(CDivTable, self).__init__()
        self.gradInput = []

    def updateOutput(self, input):
        self.output.resize_as_(input[0]).copy_(input[0])
        self.output.div_(input[1])
        return self.output

    def updateGradInput(self, input, gradOutput):
        while len(self.gradInput) < 2:
            self.gradInput.append(input[0].new())
        self.gradInput[0].resize_as_(input[0]).copy_(gradOutput, broadcast=False).div_(input[1])
        self.gradInput[1].resize_as_(input[1]).zero_().addcdiv_(-1, self.gradInput[0], input[1]).mul_(input[0])

        del self.gradInput[len(input):]

        return self.gradInput



import torch
from .Module import Module


class Tanh(Module):

    def updateOutput(self, input):
        self._backend.Tanh_updateOutput(
            self._backend.library_state,
            input,
            self.output
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        self._backend.Tanh_updateGradInput(
            self._backend.library_state,
            gradOutput,
            self.gradInput,
            self.output
        )
        return self.gradInput



import torch
from .Criterion import Criterion


class MarginRankingCriterion(Criterion):

    def __init__(self, margin=1, sizeAverage=True):
        super(MarginRankingCriterion, self).__init__()
        self.margin = margin
        self.sizeAverage = sizeAverage
        self.gradInput = [torch.Tensor(), torch.Tensor()]

        self._output = None
        self.dist = None
        self.mask = None

    def updateOutput(self, input, y):
        if input[0].size(0) == 1:
            self.output = max(0, -y * (input[0][0] - input[1][0]) + self.margin)
        else:
            if self._output is None:
                self._output = input[0].clone()
            self._output.resize_as_(input[0])
            self._output.copy_(input[0])

            self._output.add_(-1, input[1])
            self._output.mul_(-1).mul_(y)
            self._output.add_(self.margin)

            self._output.clamp_(min=0)

            self.output = self._output.sum()

            if self.sizeAverage:
                self.output = self.output / y.size(0)

        return self.output

    def updateGradInput(self, input, y):
        if input[0].size(0) == 1:
            dist = -y * (input[0][0] - input[1][0]) + self.margin
            if dist < 0:
                self.gradInput[0][0] = 0
                self.gradInput[1][0] = 0
            else:
                self.gradInput[0][0] = -y
                self.gradInput[1][0] = y
        else:
            if self.dist is None:
                self.dist = input[0].new()
            self.dist = self.dist.resize_as_(input[0]).copy_(input[0])
            dist = self.dist

            dist.add_(-1, input[1])
            dist.mul_(-1).mul_(y)
            dist.add_(self.margin)

            if self.mask is None:
                self.mask = input[0].new()
            self.mask = self.mask.resize_as_(input[0]).copy_(dist)
            mask = self.mask

            torch.ge(dist, 0, out=mask)

            self.gradInput[0].resize_(dist.size())
            self.gradInput[1].resize_(dist.size())

            self.gradInput[0].copy_(mask)
            self.gradInput[0].mul_(-1).mul_(y)
            self.gradInput[1].copy_(mask)
            self.gradInput[1].mul_(y)

            if self.sizeAverage:
                self.gradInput[0].div_(y.size(0))
                self.gradInput[1].div_(y.size(0))

        return self.gradInput



import torch

import hashlib
import os
import re
import shutil
import sys
import tempfile

try:
    from requests.utils import urlparse
    import requests.get as urlopen
    requests_available = True
except ImportError:
    requests_available = False
    if sys.version_info[0] == 2:
        from urlparse import urlparse  # noqa f811
        from urllib2 import urlopen  # noqa f811
    else:
        from urllib.request import urlopen
        from urllib.parse import urlparse
try:
    from tqdm import tqdm
except ImportError:
    tqdm = None  # defined below

# matches bfd8deac from resnet18-bfd8deac.pth
HASH_REGEX = re.compile(r'-([a-f0-9]*)\.')


def load_url(url, model_dir=None, map_location=None):
    r"""Loads the Torch serialized object at the given URL.

    If the object is already present in `model_dir`, it's deserialized and
    returned. The filename part of the URL should follow the naming convention
    ``filename-<sha256>.ext`` where ``<sha256>`` is the first eight or more
    digits of the SHA256 hash of the contents of the file. The hash is used to
    ensure unique names and to verify the contents of the file.

    The default value of `model_dir` is ``$TORCH_HOME/models`` where
    ``$TORCH_HOME`` defaults to ``~/.torch``. The default directory can be
    overriden with the ``$TORCH_MODEL_ZOO`` environment variable.

    Args:
        url (string): URL of the object to download
        model_dir (string, optional): directory in which to save the object
        map_location (optional): a function or a dict specifying how to remap storage locations (see torch.load)

    Example:
        >>> state_dict = torch.utils.model_zoo.load_url('https://s3.amazonaws.com/pytorch/models/resnet18-5c106cde.pth')

    """
    if model_dir is None:
        torch_home = os.path.expanduser(os.getenv('TORCH_HOME', '~/.torch'))
        model_dir = os.getenv('TORCH_MODEL_ZOO', os.path.join(torch_home, 'models'))
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    parts = urlparse(url)
    filename = os.path.basename(parts.path)
    cached_file = os.path.join(model_dir, filename)
    if not os.path.exists(cached_file):
        sys.stderr.write('Downloading: "{}" to {}\n'.format(url, cached_file))
        hash_prefix = HASH_REGEX.search(filename).group(1)
        _download_url_to_file(url, cached_file, hash_prefix)
    return torch.load(cached_file, map_location=map_location)


def _download_url_to_file(url, dst, hash_prefix):
    u = urlopen(url)
    if requests_available:
        file_size = int(u.headers["Content-Length"])
        u = u.raw
    else:
        meta = u.info()
        if hasattr(meta, 'getheaders'):
            file_size = int(meta.getheaders("Content-Length")[0])
        else:
            file_size = int(meta.get_all("Content-Length")[0])

    f = tempfile.NamedTemporaryFile(delete=False)
    try:
        sha256 = hashlib.sha256()
        with tqdm(total=file_size) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                sha256.update(buffer)
                pbar.update(len(buffer))

        f.close()
        digest = sha256.hexdigest()
        if digest[:len(hash_prefix)] != hash_prefix:
            raise RuntimeError('invalid hash value (expected "{}", got "{}")'
                               .format(hash_prefix, digest))
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


if tqdm is None:
    # fake tqdm if it's not installed
    class tqdm(object):

        def __init__(self, total):
            self.total = total
            self.n = 0

        def update(self, n):
            self.n += n
            sys.stderr.write("\r{0:.1f}%".format(100 * self.n / float(self.total)))
            sys.stderr.flush()

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            sys.stderr.write('\n')



import torch
from .Criterion import Criterion


class MSECriterion(Criterion):

    def __init__(self, sizeAverage=True):
        super(MSECriterion, self).__init__()
        self.sizeAverage = sizeAverage
        self.output_tensor = None

    def updateOutput(self, input, target):
        if self.output_tensor is None:
            self.output_tensor = input.new(1)
        self._backend.MSECriterion_updateOutput(
            self._backend.library_state,
            input,
            target,
            self.output_tensor,
            self.sizeAverage,
            True,  # reduce
        )
        self.output = self.output_tensor[0]
        return self.output

    def updateGradInput(self, input, target):
        implicit_gradOutput = torch.Tensor([1]).type(input.type())

        self._backend.MSECriterion_updateGradInput(
            self._backend.library_state,
            input,
            target,
            implicit_gradOutput,
            self.gradInput,
            self.sizeAverage,
            True,  # reduce
        )
        return self.gradInput



import torch
from torch.autograd._functions.utils import check_onnx_broadcast  # TODO: move me
from torch.nn.modules.utils import _pair
import warnings

# EDITING THIS FILE? READ THIS FIRST!
#
# - Parameter ordering does NOT necessarily match what is in VariableType.cpp;
#   tensors are always first, then non-tensor arguments.
# - Parameter names must *exactly* match the names in VariableType.cpp, because
#   dispatch is done with keyword arguments.
# - Looking for inplace ops?  They're detected by the trailing underscore, and
#   transparently dispatched to their non inplace versions in
#   'run_symbolic_function'.   See Note [Export inplace]

# ---------------------------------------------------------------------
# Helper functions
# ---------------------------------------------------------------------


def _scalar(x):
    """Convert a scalar tensor into a Python value."""
    assert x.numel() == 1
    return x[0]


def _if_scalar_type_as(self, tensor):
    """
    Convert self into the same type of tensor, as necessary.

    We only support implicit casting for scalars, so we never
    actually need to insert an ONNX cast operator here; just
    fix up the scalar.
    """
    if isinstance(self, torch._C.Node):
        return self
    else:
        ty = tensor.type().scalarType().lower()
        return getattr(self, ty)()


def _broadcast_if_scalar(x):
    """Return kwargs enabling broadcasting if 'x' is a scalar."""
    if isinstance(x, torch._C.Node):
        return {}
    else:
        return {"broadcast_i": 1}


def _unimplemented(op, msg):
    warnings.warn("ONNX export failed on " + op + " because " + msg + " not supported")


# ---------------------------------------------------------------------
# ONNX operator version
# ---------------------------------------------------------------------

# READ ME BEFORE EDITING _onnx_opset_version:
#
# The variable below controls which ONNX operator set version we are
# targeting.   THIS VARIABLE HAS SEMANTIC EFFECT!  Say a breaking
# change occurred in version 8.  As long as this variable < 8, you can
# export models targeting the old behavior.  However, if you bump
# this variable to 8 or later, the breaking change will take into effect:
# you MUST adjust any symbolic affected by breaking changes.  The ONNX
# spec publishes a *comprehensive* list of BC-breaking changes for every
# operator revision at:
#
#   https://github.com/onnx/onnx/blob/master/docs/Changelog.md
#
# Please be sure to go through and check all of our implementations here before
# increasing this number.  This includes symbolic definitions NOT in this
# file, so grep for "OpName" (with quotes)

_onnx_opset_version = 2


# ---------------------------------------------------------------------
# Symbolic definitions
# ---------------------------------------------------------------------


# Note [Pointwise by scalar]
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# What happens if you add a tensor with a constant (e.g., x + 2)?  There are
# some moving parts to implementing the ONNX translation in this case:
#
#   - By the time we get the scalar in a symbolic function here, it is no longer
#     a Python long/float, but a PyTorch tensor with numel == 1 (eventually, we
#     want it to be a zero dim tensor but this change has not happened yet.)
#     However, the type of this scalar is *exactly* what the user wrote in
#     Python, which may not match the tensor it is being added to.  PyTorch
#     will do implicit conversions on scalars; however, ONNX will not, so
#     we must do the conversion ourselves.  This is what _if_scalar_type_as
#     does.
#
#   - Most of the time, the arguments to self/other are pre-expanded according
#     to broadcasting.  However, a scalar will NOT be broadcasted, so we have
#     to enable broadcasting ONNX side.
#
#   - Dispatch to these functions takes advantage an outrageous coincidence
#     between the tensor and scalar name.  When we add two tensors together,
#     you get the dispatch:
#
#       add(*[self, other], **{"alpha": alpha})
#
#     When you add a tensor and a scalar, you get the dispatch:
#
#       add(*[self], **{"other": other, "alpha": alpha})
#
#     By having the argument name line up with the name of the scalar attribute
#     if it exists, we can write a single function for both overloads.
#


def add(g, self, other, alpha):
    if _scalar(alpha) != 1:
        return _unimplemented("add", "alpha != 1")
    # See Note [Pointwise by scalar]
    return g.op("Add", self, _if_scalar_type_as(other, self), **_broadcast_if_scalar(other))


def sub(g, self, other, alpha):
    if _scalar(alpha) != 1:
        return _unimplemented("sub", "alpha != 1")
    # See Note [Pointwise by scalar]
    return g.op("Sub", self, _if_scalar_type_as(other, self), **_broadcast_if_scalar(other))


def mul(g, self, other):
    # See Note [Pointwise by scalar]
    return g.op("Mul", self, _if_scalar_type_as(other, self), **_broadcast_if_scalar(other))


def div(g, self, other):
    # See Note [Pointwise by scalar]
    return g.op("Div", self, _if_scalar_type_as(other, self), **_broadcast_if_scalar(other))


# This syntax is Python 2 portable
def cat(g, *tensors, **kwargs):
    dim = kwargs.pop("dim")
    assert not kwargs
    return g.op("Concat", *tensors, axis_i=dim)


def mm(g, self, other):
    # Create a dummy C tensor. Only needed for API purposes, the value is
    # since beta = 0
    ty = self.type().scalarType().lower()
    C = g.constant(0, [1], ty)
    return g.op("Gemm", self, other, C, beta_f=0.0, alpha_f=1.0, broadcast_i=True)


def bmm(g, self, other):
    return g.op("MatMul", self, other)


def addmm(g, self, mat1, mat2, beta, alpha):
    return g.op("Gemm", mat1, mat2, self, beta_f=_scalar(beta), alpha_f=_scalar(alpha))


def neg(g, self):
    return g.op("Neg", self)


def tanh(g, self):
    return g.op("Tanh", self)


def sigmoid(g, self):
    return g.op("Sigmoid", self)


def mean(g, self, dim=None, keepdim=None):
    kwargs = {}
    # NB: ONNX's default is different from PyTorch's
    if keepdim is None:
        keepdim = 0
    return g.op("ReduceMean", self, axes_i=dim, keepdims_i=keepdim)


def t(g, self):
    return g.op("Transpose", self, perm_i=(1, 0))


def expand(g, self, size):
    # TODO: This is not a real ONNX operator at the moment
    return g.op("Expand", self, shape_i=size)


def transpose(g, self, dim0, dim1):
    if dim0 == dim1:  # micro-optimization
        return self

    # NB: Transpose in ONNX is actually a Permute
    axes = list(range(len(self.type().sizes())))
    axes[dim0], axes[dim1] = axes[dim1], axes[dim0]
    return g.op("Transpose", self, perm_i=axes)


def permute(g, self, dims):
    if dims == list(range(0, len(dims))):
        return self
    return g.op("Transpose", self, perm_i=dims)


def view(g, self, size):
    return g.op("Reshape", self, shape_i=size)


def split(g, self, split_size, dim):
    size = self.type().sizes()[dim]
    splits = [split_size] * (size // split_size)
    leftover = size % split_size
    if leftover:
        splits.append(leftover)
    return g.op("Split", self, split_i=splits, axis_i=dim, outputs=len(splits))


def squeeze(g, self, dim=None):
    if dim is None:
        dims = []
        for i, size in enumerate(self.type().sizes()):
            if size == 1:
                dims.append(i)
    else:
        dims = [dim]
    return g.op("Squeeze", self, axes_i=dims)


# NB: This appears to be dead at the moment
def prelu(g, input, weight):
    if all(s == 1 for s in weight.type().sizes()):
        return _unimplemented("prelu", "single weight shared among input channels")
    return g.op("PRelu", input, weight)


def threshold(g, input, threshold, value, inplace=False):
    # See Note [Export inplace]
    if _scalar(threshold) != 0:
        return _unimplemented("threshold", "non-zero threshold")
    if _scalar(value) != 0:
        return _unimplemented("threshold", "non-zero value")
    return g.op("Relu", input)


def leaky_relu(g, input, negative_slope, inplace=False):
    # See Note [Export inplace]
    # TODO: Talk to ONNX about unconditional cast of scalar to float
    return g.op("LeakyRelu", input, alpha_f=_scalar(negative_slope))


def glu(g, input, dim):
    assert input.type().sizes()[dim] % 2 == 0

    first, second = g.op('Split', input, axis_i=dim, outputs=2)
    return g.op('Mul', first, g.op('Sigmoid', second))


def softmax(g, input, dim=None):
    return g.op('Softmax', input, axis_i=dim)


def max_pool2d(g, input, kernel_size, stride, padding, dilation, ceil_mode):
    if ceil_mode:
        return _unimplemented("max_pool2d", "ceil_mode")
    if set(_pair(dilation)) != {1}:
        return _unimplemented("max_pool2d", "dilation")
    if not stride:
        stride = kernel_size
    r = g.op("MaxPool", input,
             kernel_shape_i=_pair(kernel_size),
             pads_i=_pair(padding),
             strides_i=_pair(stride))
    return r, None


def avg_pool2d(g, input, kernel_size, stride, padding, ceil_mode, count_include_pad):
    if ceil_mode:
        return _unimplemented("avg_pool2d", "ceil_mode")
    if not stride:
        stride = kernel_size
    # TODO: What about count_include_pad?!
    return g.op("AveragePool", input,
                kernel_shape_i=_pair(kernel_size),
                strides_i=_pair(stride),
                pads_i=_pair(padding))


def log_softmax(g, input, dim=None):
    return g.op("Log", g.op('Softmax', input, axis_i=dim).setTypeAs(input))


def unfold(g, input, dimension, size, step):
    return g.op("ATen", input, operator_s="unfold", dimension_i=dimension, size_i=size, step_i=step)


def elu(g, input, alpha, inplace=False):
    # See Note [Export inplace]
    return g.op("Elu", input, alpha_f=_scalar(alpha))



"""
This package adds support for CUDA tensor types, that implement the same
function as CPU tensors, but they utilize GPUs for computation.

It is lazily initialized, so you can always import it, and use
:func:`is_available()` to determine if your system supports CUDA.

:ref:`cuda-semantics` has more details about working with CUDA.
"""

import contextlib
import platform
import ctypes
import os
import torch
import traceback
import warnings
from torch._six import raise_from
from multiprocessing.util import register_after_fork as _register_after_fork

_initialized = False
_queued_calls = []  # don't invoke these until initialization occurs
_in_bad_fork = False  # this global is also used in torch.manual_seed
_original_pid = False
_cudart = None


def is_available():
    """Returns a bool indicating if CUDA is currently available."""
    if (not hasattr(torch._C, '_cuda_isDriverSufficient') or
            not torch._C._cuda_isDriverSufficient()):
        return False
    return torch._C._cuda_getDeviceCount() > 0


def _sleep(cycles):
    torch._C._cuda_sleep(cycles)


def _load_cudart():
    # First check the main program for CUDA symbols
    lib = ctypes.cdll.LoadLibrary(None)
    if hasattr(lib, 'cudaGetErrorName'):
        return lib

    raise RuntimeError(
        "couldn't find libcudart. Make sure CUDA libraries are installed in a"
        "default location, or that they're in {}."
        .format('DYLD_LIBRARY_PATH' if platform.system() == 'Darwin' else
                'LD_LIBRARY_PATH'))


def _check_driver():
    if not hasattr(torch._C, '_cuda_isDriverSufficient'):
        raise AssertionError("Torch not compiled with CUDA enabled")
    if not torch._C._cuda_isDriverSufficient():
        if torch._C._cuda_getDriverVersion() == 0:
            # found no NVIDIA driver on the system
            raise AssertionError("""
Found no NVIDIA driver on your system. Please check that you
have an NVIDIA GPU and installed a driver from
http://www.nvidia.com/Download/index.aspx""")
        else:
            # TODO: directly link to the alternative bin that needs install
            raise AssertionError("""
The NVIDIA driver on your system is too old (found version {}).
Please update your GPU driver by downloading and installing a new
version from the URL: http://www.nvidia.com/Download/index.aspx
Alternatively, go to: http://pytorch.org to install
a PyTorch version that has been compiled with your version
of the CUDA driver.""".format(str(torch._C._cuda_getDriverVersion())))


def _check_capability():
    incorrect_binary_warn = """
    Found GPU%d %s which requires CUDA_VERSION >= %d for
     optimal performance and fast startup time, but your PyTorch was compiled
     with CUDA_VERSION %d. Please install the correct PyTorch binary
     using instructions from http://pytorch.org
    """

    old_gpu_warn = """
    Found GPU%d %s which is of cuda capability %d.%d.
    PyTorch no longer supports this GPU because it is too old.
    """

    CUDA_VERSION = torch._C._cuda_getCompiledVersion()
    for d in range(device_count()):
        capability = get_device_capability(d)
        major = capability[0]
        name = get_device_name(d)
        if CUDA_VERSION < 8000 and major >= 6:
            warnings.warn(incorrect_binary_warn % (d, name, 8000, CUDA_VERSION))
        elif CUDA_VERSION < 9000 and major >= 7:
            warnings.warn(incorrect_binary_warn % (d, name, 9000, CUDA_VERSION))
        elif capability == (3, 0) or capability == (5, 0) or major < 3:
            warnings.warn(old_gpu_warn % (d, name, major, capability[1]))


def _lazy_call(callable):
    if _initialized:
        callable()
    else:
        # Don't store the actual traceback to avoid memory cycle
        _queued_calls.append((callable, traceback.format_stack()))

_lazy_call(_check_capability)


class DeferredCudaCallError(Exception):
    pass


def init():
    """Initialize PyTorch's CUDA state.  You may need to call
    this explicitly if you are interacting with PyTorch via
    its C API, as Python bindings for CUDA functionality will not
    be until this initialization takes place.  Ordinary users
    should not need this, as all of PyTorch's CUDA methods
    automatically initialize CUDA state on-demand.

    Does nothing if the CUDA state is already initialized.
    """
    _lazy_init()


def _lazy_init():
    global _initialized, _cudart, _original_pid, _queued_calls
    if _initialized:
        return
    if _in_bad_fork:
        from sys import version_info
        if version_info < (3, 4):
            msg = ("To use CUDA with multiprocessing, you must use Python "
                   "3.4+ and the 'spawn' start method")
        else:
            msg = ("To use CUDA with multiprocessing, you must use the "
                   "'spawn' start method")
        raise RuntimeError(
            "Cannot re-initialize CUDA in forked subprocess. " + msg)
    _check_driver()
    torch._C._cuda_init()
    torch._C._cuda_sparse_init()
    _cudart = _load_cudart()
    _cudart.cudaGetErrorName.restype = ctypes.c_char_p
    _cudart.cudaGetErrorString.restype = ctypes.c_char_p
    _original_pid = os.getpid()
    _initialized = True
    # Important to do this after _initialized, since some queued calls
    # may themselves call _lazy_init()
    for queued_call, orig_traceback in _queued_calls:
        try:
            queued_call()
        except Exception as e:
            msg = ("CUDA call failed lazily at initialization with error: {}\n\n"
                   "CUDA call was originally invoked at:\n\n{}").format(str(e), orig_traceback)
            raise_from(DeferredCudaCallError(msg), e)


def _after_fork(arg):
    global _initialized, _in_bad_fork
    if _initialized and _original_pid != os.getpid():
        _initialized = False
        _in_bad_fork = True
        _CudaBase.__new__ = _lazy_new


_register_after_fork(_after_fork, _after_fork)


def cudart():
    _lazy_init()
    return _cudart


class cudaStatus(object):
    SUCCESS = 0
    ERROR_NOT_READY = 34


class CudaError(RuntimeError):
    def __init__(self, code):
        msg = cudart().cudaGetErrorString(code).decode('utf-8')
        super(CudaError, self).__init__('{0} ({1})'.format(msg, code))


def check_error(res):
    if res != cudaStatus.SUCCESS:
        raise CudaError(res)


class device(object):
    """Context-manager that changes the selected device.

    Arguments:
        idx (int): device index to select. It's a no-op if this argument
            is negative.
    """

    def __init__(self, idx):
        self.idx = idx
        self.prev_idx = -1

    def __enter__(self):
        if self.idx is -1:
            return
        self.prev_idx = torch._C._cuda_getDevice()
        if self.prev_idx != self.idx:
            torch._C._cuda_setDevice(self.idx)
        _lazy_init()

    def __exit__(self, *args):
        if self.prev_idx != self.idx:
            torch._C._cuda_setDevice(self.prev_idx)
        return False


class device_of(device):
    """Context-manager that changes the current device to that of given object.

    You can use both tensors and storages as arguments. If a given object is
    not allocated on a GPU, this is a no-op.

    Arguments:
        obj (Tensor or Storage): object allocated on the selected device.
    """

    def __init__(self, obj):
        idx = obj.get_device() if obj.is_cuda else -1
        super(device_of, self).__init__(idx)


def set_device(device):
    """Sets the current device.

    Usage of this function is discouraged in favor of :any:`device`. In most
    cases it's better to use ``CUDA_VISIBLE_DEVICES`` environmental variable.

    Arguments:
        device (int): selected device. This function is a no-op if this
            argument is negative.
    """
    if device >= 0:
        torch._C._cuda_setDevice(device)


def get_device_name(device):
    """Gets the name of a device.

    Arguments:
        device (int): device for which to return the name. This function is a
            no-op if this argument is negative.
    """
    if device >= 0:
        return torch._C._cuda_getDeviceName(device)


def get_device_capability(device):
    """Gets the cuda capability of a device.

    Arguments:
        device (int): device for which to return the name. This function is a
            no-op if this argument is negative.
    Returns:
        tuple(int, int): the major and minor cuda capability of the device
    """
    if device >= 0:
        return torch._C._cuda_getDeviceCapability(device)


@contextlib.contextmanager
def stream(stream):
    """Context-manager that selects a given stream.

    All CUDA kernels queued within its context will be enqueued on a selected
    stream.

    Arguments:
        stream (Stream): selected stream. This manager is a no-op if it's
            ``None``.

    .. note:: Streams are per-device, and this function changes the "current
       stream" only for the currently selected device.  It is illegal to select
       a stream that belongs to a different device.
    """
    if stream is None:
        yield
        return
    prev_stream = current_stream()
    torch._C._cuda_setStream(stream._cdata)
    try:
        yield
    finally:
        torch._C._cuda_setStream(prev_stream._cdata)


def device_count():
    """Returns the number of GPUs available."""
    if is_available():
        return torch._C._cuda_getDeviceCount()
    else:
        return 0


def current_device():
    """Returns the index of a currently selected device."""
    _lazy_init()
    return torch._C._cuda_getDevice()


def synchronize():
    """Waits for all kernels in all streams on current device to complete."""
    _lazy_init()
    return torch._C._cuda_synchronize()


def current_stream():
    """Returns a currently selected :class:`Stream`."""
    _lazy_init()
    return torch.cuda.Stream(_cdata=torch._C._cuda_getCurrentStream())


def current_blas_handle():
    """Returns cublasHandle_t pointer to current cuBLAS handle"""
    _lazy_init()
    return torch._C._cuda_getCurrentBlasHandle()


def empty_cache():
    """Releases all unoccupied cached memory currently held by the caching
    allocator so that those can be used in other GPU application and visible in
    `nvidia-smi`."""
    if _initialized:
        return torch._C._cuda_emptyCache()


def _host_allocator():
    _lazy_init()
    return torch._C._cuda_cudaHostAllocator()


@contextlib.contextmanager
def _free_mutex():
    torch._C._cuda_lock_mutex()
    try:
        yield
    finally:
        torch._C._cuda_unlock_mutex()


from .random import *

################################################################################
# Define Storage and Tensor classes
################################################################################


from ..tensor import _TensorBase
from ..storage import _StorageBase


def _dummy_type(name):
    def init_err(self):
        class_name = self.__class__.__name__
        raise RuntimeError(
            "Tried to instantiate dummy base class {}".format(class_name))
    return type(storage_name, (object,), {"__init__": init_err})


if not hasattr(torch._C, 'CudaDoubleStorageBase'):
    # Define dummy base classes
    for t in ['Double', 'Float', 'Long', 'Int', 'Short', 'Char', 'Byte', 'Half']:
        storage_name = 'Cuda{0}StorageBase'.format(t)
        tensor_name = 'Cuda{0}TensorBase'.format(t)

        torch._C.__dict__[storage_name] = _dummy_type(storage_name)
        torch._C.__dict__[tensor_name] = _dummy_type(tensor_name)

    torch._C.__dict__['_CudaStreamBase'] = _dummy_type('CudaStreamBase')


@staticmethod
def _lazy_new(cls, *args, **kwargs):
    _lazy_init()
    # We need this method only for lazy init, so we can remove it
    del _CudaBase.__new__
    return super(_CudaBase, cls).__new__(cls, *args, **kwargs)


class _CudaBase(object):
    is_cuda = True
    is_sparse = False

    def type(self, *args, **kwargs):
        with device(self.get_device()):
            return super(_CudaBase, self).type(*args, **kwargs)

    __new__ = _lazy_new


class DoubleStorage(_CudaBase, torch._C.CudaDoubleStorageBase, _StorageBase):
    pass


class FloatStorage(_CudaBase, torch._C.CudaFloatStorageBase, _StorageBase):
    pass


class LongStorage(_CudaBase, torch._C.CudaLongStorageBase, _StorageBase):
    pass


class IntStorage(_CudaBase, torch._C.CudaIntStorageBase, _StorageBase):
    pass


class ShortStorage(_CudaBase, torch._C.CudaShortStorageBase, _StorageBase):
    pass


class CharStorage(_CudaBase, torch._C.CudaCharStorageBase, _StorageBase):
    pass


class ByteStorage(_CudaBase, torch._C.CudaByteStorageBase, _StorageBase):
    pass


class HalfStorage(_CudaBase, torch._C.CudaHalfStorageBase, _StorageBase):
    pass


class DoubleTensor(_CudaBase, torch._C.CudaDoubleTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type(cls):
        return DoubleStorage


class FloatTensor(_CudaBase, torch._C.CudaFloatTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type(cls):
        return FloatStorage


class LongTensor(_CudaBase, torch._C.CudaLongTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type(cls):
        return LongStorage


class IntTensor(_CudaBase, torch._C.CudaIntTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type(cls):
        return IntStorage


class ShortTensor(_CudaBase, torch._C.CudaShortTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type(cls):
        return ShortStorage


class CharTensor(_CudaBase, torch._C.CudaCharTensorBase, _TensorBase):

    def is_signed(self):
        # TODO
        return False

    @classmethod
    def storage_type(cls):
        return CharStorage


class ByteTensor(_CudaBase, torch._C.CudaByteTensorBase, _TensorBase):

    def is_signed(self):
        return False

    @classmethod
    def storage_type(cls):
        return ByteStorage


class HalfTensor(_CudaBase, torch._C.CudaHalfTensorBase, _TensorBase):

    def is_signed(self):
        return True

    @classmethod
    def storage_type():
        return HalfStorage


torch._storage_classes.add(DoubleStorage)
torch._storage_classes.add(FloatStorage)
torch._storage_classes.add(LongStorage)
torch._storage_classes.add(IntStorage)
torch._storage_classes.add(ShortStorage)
torch._storage_classes.add(CharStorage)
torch._storage_classes.add(ByteStorage)
torch._storage_classes.add(HalfStorage)

torch._tensor_classes.add(DoubleTensor)
torch._tensor_classes.add(FloatTensor)
torch._tensor_classes.add(LongTensor)
torch._tensor_classes.add(IntTensor)
torch._tensor_classes.add(ShortTensor)
torch._tensor_classes.add(CharTensor)
torch._tensor_classes.add(ByteTensor)
torch._tensor_classes.add(HalfTensor)

from . import sparse
from . import profiler
from . import nvtx
from .streams import Stream, Event



import torch
from .Module import Module


class MulConstant(Module):

    def __init__(self, constant_scalar, inplace=False):
        super(MulConstant, self).__init__()
        self.constant_scalar = constant_scalar
        self.inplace = inplace

    def updateOutput(self, input):
        if self.inplace:
            input.mul_(self.constant_scalar)
            self.output.set_(input)
        else:
            self.output.resize_as_(input)
            self.output.copy_(input)
            self.output.mul_(self.constant_scalar)

        return self.output

    def updateGradInput(self, input, gradOutput):
        if self.gradInput is None:
            return

        if self.inplace:
            gradOutput.mul_(self.constant_scalar)
            self.gradInput.set_(gradOutput)
            # restore previous input value
            input.div_(self.constant_scalar)
        else:
            self.gradInput.resize_as_(gradOutput)
            self.gradInput.copy_(gradOutput)
            self.gradInput.mul_(self.constant_scalar)

        return self.gradInput



import torch
from numbers import Number
from .function import Function

_NOT_PROVIDED = object()


class StochasticFunction(Function):

    def __init__(self):
        self.reward = _NOT_PROVIDED

    def _do_backward(self, grad_output, retain_variables):
        if self.reward is _NOT_PROVIDED:
            raise RuntimeError("differentiating stochastic functions requires "
                               "providing a reward")
        result = super(StochasticFunction, self)._do_backward((self.reward,), retain_variables)
        if not retain_variables:
            self.reward = None
        return result

    def _do_forward(self, *inputs):
        result = super(StochasticFunction, self)._do_forward(*inputs)
        # save output type and size, to check the type of reward
        assert isinstance(result, torch.autograd.Variable), \
            "stochastic functions support only a single output at the moment"
        self.reward_info = (type(inputs[0].data), result.size())
        return result

    __call__ = _do_forward

    def _reinforce(self, reward):
        is_number = isinstance(reward, Number)
        if not is_number and type(reward) != self.reward_info[0]:
            raise TypeError("mismatch between reward and output type: got {}, "
                            "but expected {}".format(torch.typename(reward),
                                                     torch.typename(self.reward_info[0])))
        if not is_number and reward.size() != self.reward_info[1]:
            raise ValueError("got reward of size {}, but expected a tensor of size {}".format(
                             'x'.join(map(str, reward.size())),
                             'x'.join(map(str, self.reward_info[1]))))
        if self.reward is not _NOT_PROVIDED:
            raise RuntimeError("you can only reinforce a stochastic Function once")
        self.reward = reward



import ctypes

lib = None

__all__ = ['range_push', 'range_pop', 'mark']


def _libnvToolsExt():
    global lib
    if lib is None:
        lib = ctypes.cdll.LoadLibrary(None)
        lib.nvtxMarkA.restype = None
    return lib


def range_push(msg):
    """
    Pushes a range onto a stack of nested range span.  Returns zero-based
    depth of the range that is started.

    Arguments:
        msg (string): ASCII message to associate with range
    """
    if _libnvToolsExt() is None:
        raise RuntimeError('Unable to load nvToolsExt library')
    return lib.nvtxRangePushA(ctypes.c_char_p(msg.encode("ascii")))


def range_pop():
    """
    Pops a range off of a stack of nested range spans.  Returns the
    zero-based depth of the range that is ended.
    """
    if _libnvToolsExt() is None:
        raise RuntimeError('Unable to load nvToolsExt library')
    return lib.nvtxRangePop()


def mark(msg):
    """
    Describe an instantaneous event that occurred at some point.

    Arguments:
        msg (string): ASCII message to associate with the event.
    """
    if _libnvToolsExt() is None:
        raise RuntimeError('Unable to load nvToolsExt library')
    return lib.nvtxMarkA(ctypes.c_char_p(msg.encode("ascii")))



import torch
from .Module import Module
from .utils import clear


class SpatialAdaptiveMaxPooling(Module):

    def __init__(self, w, h):
        super(SpatialAdaptiveMaxPooling, self).__init__()
        self.w = w
        self.h = h
        self.indices = None

    def updateOutput(self, input):
        if self.indices is None:
            self.indices = input.new()
        self.indices = self.indices.long()
        self._backend.SpatialAdaptiveMaxPooling_updateOutput(
            self._backend.library_state,
            input,
            self.output,
            self.indices,
            self.w,
            self.h
        )
        return self.output

    def updateGradInput(self, input, gradOutput):
        self._backend.SpatialAdaptiveMaxPooling_updateGradInput(
            self._backend.library_state,
            input,
            gradOutput,
            self.gradInput,
            self.indices
        )
        return self.gradInput

    def clearState(self):
        clear(self, 'indices')
        return super(SpatialAdaptiveMaxPooling, self).clearState()
